@@ -168,8 +168,12 @@ static int impl_node_set_param(void *object, uint32_t id, uint32_t flags,
 
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 
+
 	switch (id) {
 	case SPA_PARAM_Props:
+	{
+		struct port *port = &this->port;
+
 		if (param == NULL) {
 			reset_props(this, &this->props);
 			return 0;
@@ -179,11 +183,11 @@ static int impl_node_set_param(void *object, uint32_t id, uint32_t flags,
 			SPA_PROP_live, SPA_POD_OPT_Bool(&this->props.live));
 
 		if (this->props.live)
-			this->info.flags |= SPA_PORT_FLAG_LIVE;
+			port->info.flags |= SPA_PORT_FLAG_LIVE;
 		else
-			this->info.flags &= ~SPA_PORT_FLAG_LIVE;
+			port->info.flags &= ~SPA_PORT_FLAG_LIVE;
 		break;
-
+	}
 	default:
 		return -ENOENT;
 	}
@@ -482,8 +486,7 @@ impl_node_port_enum_params(void *object, int seq,
 			SPA_PARAM_BUFFERS_buffers, SPA_POD_CHOICE_RANGE_Int(2, 1, 32),
 			SPA_PARAM_BUFFERS_blocks,  SPA_POD_Int(1),
 			SPA_PARAM_BUFFERS_size,    SPA_POD_Int(128),
-			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(1),
-			SPA_PARAM_BUFFERS_align,   SPA_POD_Int(16));
+			SPA_PARAM_BUFFERS_stride,  SPA_POD_Int(1));
 		break;
 	case SPA_PARAM_Meta:
 		switch (result.index) {
@@ -699,6 +702,13 @@ static int impl_get_interface(struct spa_handle *handle, const char *type, void
 	return 0;
 }
 
+static int do_remove_timer(struct spa_loop *loop, bool async, uint32_t seq, const void *data, size_t size, void *user_data)
+{
+	struct impl *this = user_data;
+	spa_loop_remove_source(this->data_loop, &this->timer_source);
+	return 0;
+}
+
 static int impl_clear(struct spa_handle *handle)
 {
 	struct impl *this;
@@ -708,7 +718,7 @@ static int impl_clear(struct spa_handle *handle)
 	this = (struct impl *) handle;
 
 	if (this->data_loop)
-		spa_loop_remove_source(this->data_loop, &this->timer_source);
+		spa_loop_invoke(this->data_loop, do_remove_timer, 0, NULL, 0, true, this);
 	spa_system_close(this->data_system, this->timer_source.fd);
 
 	return 0;
