@@ -67,7 +67,9 @@ enum
 {
   PROP_0,
   PROP_PATH,
+  PROP_TARGET_OBJECT,
   PROP_CLIENT_NAME,
+  PROP_CLIENT_PROPERTIES,
   PROP_STREAM_PROPERTIES,
   PROP_ALWAYS_COPY,
   PROP_MIN_BUFFERS,
@@ -116,15 +118,27 @@ gst_pipewire_src_set_property (GObject * object, guint prop_id,
       pwsrc->path = g_value_dup_string (value);
       break;
 
+    case PROP_TARGET_OBJECT:
+      g_free (pwsrc->target_object);
+      pwsrc->target_object = g_value_dup_string (value);
+      break;
+
     case PROP_CLIENT_NAME:
       g_free (pwsrc->client_name);
       pwsrc->client_name = g_value_dup_string (value);
       break;
 
+    case PROP_CLIENT_PROPERTIES:
+      if (pwsrc->client_properties)
+        gst_structure_free (pwsrc->client_properties);
+      pwsrc->client_properties =
+          gst_structure_copy (gst_value_get_structure (value));
+      break;
+
     case PROP_STREAM_PROPERTIES:
-      if (pwsrc->properties)
-        gst_structure_free (pwsrc->properties);
-      pwsrc->properties =
+      if (pwsrc->stream_properties)
+        gst_structure_free (pwsrc->stream_properties);
+      pwsrc->stream_properties =
           gst_structure_copy (gst_value_get_structure (value));
       break;
 
@@ -169,12 +183,20 @@ gst_pipewire_src_get_property (GObject * object, guint prop_id,
       g_value_set_string (value, pwsrc->path);
       break;
 
+    case PROP_TARGET_OBJECT:
+      g_value_set_string (value, pwsrc->target_object);
+      break;
+
     case PROP_CLIENT_NAME:
       g_value_set_string (value, pwsrc->client_name);
       break;
 
+    case PROP_CLIENT_PROPERTIES:
+      gst_value_set_structure (value, pwsrc->client_properties);
+      break;
+
     case PROP_STREAM_PROPERTIES:
-      gst_value_set_structure (value, pwsrc->properties);
+      gst_value_set_structure (value, pwsrc->stream_properties);
       break;
 
     case PROP_ALWAYS_COPY:
@@ -239,11 +261,14 @@ gst_pipewire_src_finalize (GObject * object)
 {
   GstPipeWireSrc *pwsrc = GST_PIPEWIRE_SRC (object);
 
-  if (pwsrc->properties)
-    gst_structure_free (pwsrc->properties);
+  if (pwsrc->stream_properties)
+    gst_structure_free (pwsrc->stream_properties);
+  if (pwsrc->client_properties)
+    gst_structure_free (pwsrc->client_properties);
   if (pwsrc->clock)
     gst_object_unref (pwsrc->clock);
   g_free (pwsrc->path);
+  g_free (pwsrc->target_object);
   g_free (pwsrc->client_name);
   g_object_unref(pwsrc->pool);
 
@@ -274,6 +299,16 @@ gst_pipewire_src_class_init (GstPipeWireSrcClass * klass)
                                                         "The source path to connect to (NULL = default)",
                                                         NULL,
                                                         G_PARAM_READWRITE |
+                                                        G_PARAM_STATIC_STRINGS |
+                                                        G_PARAM_DEPRECATED));
+
+  g_object_class_install_property (gobject_class,
+                                   PROP_TARGET_OBJECT,
+                                   g_param_spec_string ("target-object",
+                                                        "Target object",
+                                                        "The source name/serial to connect to (NULL = default)",
+                                                        NULL,
+                                                        G_PARAM_READWRITE |
                                                         G_PARAM_STATIC_STRINGS));
 
   g_object_class_install_property (gobject_class,
@@ -285,6 +320,15 @@ gst_pipewire_src_class_init (GstPipeWireSrcClass * klass)
                                                         G_PARAM_READWRITE |
                                                         G_PARAM_STATIC_STRINGS));
 
+  g_object_class_install_property (gobject_class,
+                                   PROP_CLIENT_PROPERTIES,
+                                   g_param_spec_boxed ("client-properties",
+                                                       "client properties",
+                                                       "list of PipeWire client properties",
+                                                       GST_TYPE_STRUCTURE,
+                                                       G_PARAM_READWRITE |
+                                                       G_PARAM_STATIC_STRINGS));
+
   g_object_class_install_property (gobject_class,
                                    PROP_STREAM_PROPERTIES,
                                    g_param_spec_boxed ("stream-properties",
@@ -476,14 +520,21 @@ static GstBuffer *dequeue_buffer(GstPipeWireSrc *pwsrc)
           return NULL;
 
   data = b->user_data;
-  buf = data->buf;
 
-  GST_LOG_OBJECT (pwsrc, "got new buffer %p", buf);
+  if (!GST_IS_BUFFER (data->buf)) {
+    GST_ERROR_OBJECT (pwsrc, "stream buffer %p is missing", data->buf);
+    return NULL;
+  }
 
   if (!data->queued) {
-    GST_WARNING_OBJECT (pwsrc, "buffer %p was not recycled", buf);
-    gst_buffer_ref (buf);
+    GST_ERROR_OBJECT (pwsrc, "buffer %p was not recycled", data->buf);
+    return NULL;
   }
+
+  GST_LOG_OBJECT (pwsrc, "got new buffer %p", data->buf);
+
+  buf = gst_buffer_new ();
+
   data->queued = FALSE;
   GST_BUFFER_PTS (buf) = GST_CLOCK_TIME_NONE;
   GST_BUFFER_DTS (buf) = GST_CLOCK_TIME_NONE;
@@ -509,12 +560,16 @@ static GstBuffer *dequeue_buffer(GstPipeWireSrc *pwsrc)
       meta->height = crop->region.size.height;
     }
   }
+  gst_buffer_add_parent_buffer_meta (buf, data->buf);
+  gst_buffer_unref (data->buf);
   for (i = 0; i < b->buffer->n_datas; i++) {
     struct spa_data *d = &b->buffer->datas[i];
-    GstMemory *mem = gst_buffer_peek_memory (buf, i);
-    mem->offset = SPA_MIN(d->chunk->offset, d->maxsize);
-    mem->size = SPA_MIN(d->chunk->size, d->maxsize - mem->offset);
-    mem->offset += data->offset;
+    GstMemory *pmem = gst_buffer_peek_memory (data->buf, i);
+    if (pmem) {
+      GstMemory *mem = gst_memory_share (pmem, d->chunk->offset, d->chunk->size);
+      gst_buffer_insert_memory (buf, i, mem);
+      spa_assert_se(mem->size <= mem->maxsize);
+    }
     if (d->chunk->flags & SPA_CHUNK_FLAG_CORRUPTED)
       GST_BUFFER_FLAG_SET (buf, GST_BUFFER_FLAG_CORRUPTED);
   }
@@ -583,7 +638,7 @@ gst_pipewire_src_stream_start (GstPipeWireSrc *pwsrc)
   GST_DEBUG_OBJECT (pwsrc, "doing stream start");
 
   pw_thread_loop_get_time (pwsrc->core->loop, &abstime,
-		  GST_PIPEWIRE_DEFAULT_TIMEOUT * SPA_NSEC_PER_SEC);
+                  GST_PIPEWIRE_DEFAULT_TIMEOUT * SPA_NSEC_PER_SEC);
 
   while (TRUE) {
     enum pw_stream_state state = pw_stream_get_state (pwsrc->stream, &error);
@@ -633,7 +688,7 @@ wait_started (GstPipeWireSrc *this)
   pw_thread_loop_lock (this->core->loop);
 
   pw_thread_loop_get_time (this->core->loop, &abstime,
-		  GST_PIPEWIRE_DEFAULT_TIMEOUT * SPA_NSEC_PER_SEC);
+                  GST_PIPEWIRE_DEFAULT_TIMEOUT * SPA_NSEC_PER_SEC);
 
   while (TRUE) {
     state = pw_stream_get_state (this->stream, &error);
@@ -658,7 +713,7 @@ wait_started (GstPipeWireSrc *this)
     }
   }
   GST_DEBUG_OBJECT (this, "got started signal: %s",
-		  pw_stream_state_as_string (state));
+                  pw_stream_state_as_string (state));
   pw_thread_loop_unlock (this->core->loop);
 
   return state;
@@ -675,6 +730,7 @@ gst_pipewire_src_negotiate (GstBaseSrc * basesrc)
   GPtrArray *possible;
   const char *error = NULL;
   struct timespec abstime;
+  uint32_t target_id;
 
   /* first see what is possible on our source pad */
   thiscaps = gst_pad_query_caps (GST_BASE_SRC_PAD (basesrc), NULL);
@@ -727,18 +783,40 @@ gst_pipewire_src_negotiate (GstBaseSrc * basesrc)
     }
   }
 
-  GST_DEBUG_OBJECT (basesrc, "connect capture with path %s", pwsrc->path);
+  target_id = pwsrc->path ? (uint32_t)atoi(pwsrc->path) : PW_ID_ANY;
+
+  if (pwsrc->target_object) {
+      struct spa_dict_item items[2] = {
+        SPA_DICT_ITEM_INIT(PW_KEY_TARGET_OBJECT, pwsrc->target_object),
+        SPA_DICT_ITEM_INIT(PW_KEY_NODE_TARGET, NULL),
+      };
+      struct spa_dict dict = SPA_DICT_INIT_ARRAY(items);
+      uint64_t serial;
+
+      /* If target.object is a name, set it also to node.target */
+      if (spa_atou64(pwsrc->target_object, &serial, 0)) {
+        dict.n_items = 1;
+      } else {
+        target_id = PW_ID_ANY;
+        items[1].value = pwsrc->target_object;
+      }
+
+      pw_stream_update_properties (pwsrc->stream, &dict);
+  }
+
+  GST_DEBUG_OBJECT (basesrc, "connect capture with path %s, target-object %s",
+                    pwsrc->path, pwsrc->target_object);
   pwsrc->negotiated = FALSE;
   pw_stream_connect (pwsrc->stream,
                      PW_DIRECTION_INPUT,
-                     pwsrc->path ? (uint32_t)atoi(pwsrc->path) : PW_ID_ANY,
+                     target_id,
                      PW_STREAM_FLAG_AUTOCONNECT | PW_STREAM_FLAG_DONT_RECONNECT,
                      (const struct spa_pod **)possible->pdata,
                      possible->len);
   g_ptr_array_free (possible, TRUE);
 
   pw_thread_loop_get_time (pwsrc->core->loop, &abstime,
-		  GST_PIPEWIRE_DEFAULT_TIMEOUT * SPA_NSEC_PER_SEC);
+                  GST_PIPEWIRE_DEFAULT_TIMEOUT * SPA_NSEC_PER_SEC);
 
   while (TRUE) {
     enum pw_stream_state state = pw_stream_get_state (pwsrc->stream, &error);
@@ -816,7 +894,7 @@ on_param_changed (void *data, uint32_t id,
     return;
   }
   if (pwsrc->caps)
-	  gst_caps_unref(pwsrc->caps);
+          gst_caps_unref(pwsrc->caps);
   pwsrc->caps = gst_caps_from_format (param);
   GST_DEBUG_OBJECT (pwsrc, "we got format %" GST_PTR_FORMAT, pwsrc->caps);
 
@@ -838,8 +916,8 @@ on_param_changed (void *data, uint32_t id,
         SPA_PARAM_BUFFERS_size,    SPA_POD_CHOICE_RANGE_Int(0, 0, INT32_MAX),
         SPA_PARAM_BUFFERS_stride,  SPA_POD_CHOICE_RANGE_Int(0, 0, INT32_MAX),
         SPA_PARAM_BUFFERS_dataType, SPA_POD_CHOICE_FLAGS_Int(
-						(1<<SPA_DATA_MemFd) |
-						(1<<SPA_DATA_MemPtr)));
+                                                (1<<SPA_DATA_MemFd) |
+                                                (1<<SPA_DATA_MemPtr)));
 
     params[1] = spa_pod_builder_add_object (&b,
         SPA_TYPE_OBJECT_ParamMeta, SPA_PARAM_Meta,
@@ -987,7 +1065,7 @@ gst_pipewire_src_create (GstPushSrc * psrc, GstBuffer ** buffer)
       buf = dequeue_buffer (pwsrc);
       GST_LOG_OBJECT (pwsrc, "popped buffer %p", buf);
       if (buf != NULL) {
-	if (pwsrc->resend_last || pwsrc->keepalive_time > 0)
+        if (pwsrc->resend_last || pwsrc->keepalive_time > 0)
           gst_buffer_replace (&pwsrc->last_buffer, buf);
         break;
       }
@@ -996,7 +1074,7 @@ gst_pipewire_src_create (GstPushSrc * psrc, GstBuffer ** buffer)
     if (pwsrc->keepalive_time > 0) {
       struct timespec abstime;
       pw_thread_loop_get_time(pwsrc->core->loop, &abstime,
-		      pwsrc->keepalive_time * SPA_NSEC_PER_MSEC);
+                      pwsrc->keepalive_time * SPA_NSEC_PER_MSEC);
       if (pw_thread_loop_timed_wait_full (pwsrc->core->loop, &abstime) == -ETIMEDOUT)
         timeout = TRUE;
     } else {
@@ -1132,12 +1210,17 @@ gst_pipewire_src_open (GstPipeWireSrc * pwsrc)
   pw_thread_loop_lock (pwsrc->core->loop);
 
   props = pw_properties_new (NULL, NULL);
+  if (pwsrc->client_properties) {
+    gst_structure_foreach (pwsrc->client_properties, copy_properties, props);
+    pw_core_update_properties (pwsrc->core->core, &props->dict);
+    pw_properties_clear(props);
+  }
   if (pwsrc->client_name) {
     pw_properties_set (props, PW_KEY_NODE_NAME, pwsrc->client_name);
     pw_properties_set (props, PW_KEY_NODE_DESCRIPTION, pwsrc->client_name);
   }
-  if (pwsrc->properties) {
-    gst_structure_foreach (pwsrc->properties, copy_properties, props);
+  if (pwsrc->stream_properties) {
+    gst_structure_foreach (pwsrc->stream_properties, copy_properties, props);
   }
 
   if ((pwsrc->stream = pw_stream_new (pwsrc->core->core,
