@@ -153,8 +153,6 @@ struct impl {
 	struct pw_impl_module *module;
 	struct spa_hook module_listener;
 
-	uint32_t id;
-
 	struct pw_core *core;
 	struct spa_hook core_proxy_listener;
 	struct spa_hook core_listener;
@@ -169,7 +167,6 @@ struct impl {
 	void *rec_buffer[SPA_AUDIO_MAX_CHANNELS];
 	uint32_t rec_ringsize;
 	struct spa_ringbuffer rec_ring;
-	struct spa_io_rate_match *rec_rate_match;
 
 	struct pw_stream *playback;
 	struct spa_hook playback_listener;
@@ -180,7 +177,6 @@ struct impl {
 	uint32_t play_ringsize;
 	struct spa_ringbuffer play_ring;
 	struct spa_ringbuffer play_delayed_ring;
-	struct spa_io_rate_match *play_rate_match;
 
 	void *out_buffer[SPA_AUDIO_MAX_CHANNELS];
 	uint32_t out_ringsize;
@@ -338,32 +334,25 @@ static void capture_destroy(void *d)
 	impl->capture = NULL;
 }
 
-static void capture_io_changed(void *data, uint32_t id, void *area, uint32_t size)
-{
-	struct impl *impl = data;
-
-	switch (id) {
-	case SPA_IO_RateMatch:
-		impl->rec_rate_match = area;
-		break;
-	}
-}
-
 static void capture_process(void *data)
 {
 	struct impl *impl = data;
 	struct pw_buffer *buf;
 	struct spa_data *d;
-	uint32_t i, index, size;
+	uint32_t i, index, offs, size;
 	int32_t avail;
 
 	if ((buf = pw_stream_dequeue_buffer(impl->capture)) == NULL) {
 		pw_log_debug("out of capture buffers: %m");
 		return;
 	}
+	d = &buf->buffer->datas[0];
+
+	offs = SPA_MIN(d->chunk->offset, d->maxsize);
+	size = SPA_MIN(d->chunk->size, d->maxsize - offs);
 
 	avail = spa_ringbuffer_get_write_index(&impl->rec_ring, &index);
-	size = buf->buffer->datas[0].chunk->size;
+
 	if (avail + size > impl->rec_ringsize) {
 		uint32_t rindex, drop;
 
@@ -382,7 +371,7 @@ static void capture_process(void *data)
 	 * if it has a specific requirement, else keep the block size the same
 	 * on input and output or what the resampler needs */
 	if (impl->aec_blocksize == 0) {
-		impl->aec_blocksize = SPA_MAX(size, impl->rec_rate_match->size);
+		impl->aec_blocksize = size;
 		pw_log_debug("Setting AEC block size to %u", impl->aec_blocksize);
 	}
 
@@ -390,10 +379,12 @@ static void capture_process(void *data)
 		/* captured samples, with echo from sink */
 		d = &buf->buffer->datas[i];
 
+		offs = SPA_MIN(d->chunk->offset, d->maxsize);
+		size = SPA_MIN(d->chunk->size, d->maxsize - offs);
+
 		spa_ringbuffer_write_data(&impl->rec_ring, impl->rec_buffer[i],
 				impl->rec_ringsize, index % impl->rec_ringsize,
-				SPA_PTROFF(d->data, d->chunk->offset, void),
-				d->chunk->size);
+				SPA_PTROFF(d->data, offs, void), size);
 	}
 
 	spa_ringbuffer_write_update(&impl->rec_ring, index + size);
@@ -416,6 +407,13 @@ static void input_state_changed(void *data, enum pw_stream_state old,
 		pw_stream_flush(impl->source, false);
 		pw_stream_flush(impl->capture, false);
 		break;
+	case PW_STREAM_STATE_UNCONNECTED:
+		pw_log_info("%p: input unconnected", impl);
+		pw_impl_module_schedule_destroy(impl->module);
+		break;
+	case PW_STREAM_STATE_ERROR:
+		pw_log_info("%p: input error: %s", impl, error);
+		break;
 	default:
 		break;
 	}
@@ -454,7 +452,6 @@ static const struct pw_stream_events capture_events = {
 	PW_VERSION_STREAM_EVENTS,
 	.destroy = capture_destroy,
 	.state_changed = input_state_changed,
-	.io_changed = capture_io_changed,
 	.process = capture_process,
 	.param_changed = input_param_changed
 };
@@ -482,6 +479,13 @@ static void output_state_changed(void *data, enum pw_stream_state old,
 		pw_stream_flush(impl->sink, false);
 		pw_stream_flush(impl->playback, false);
 		break;
+	case PW_STREAM_STATE_UNCONNECTED:
+		pw_log_info("%p: output unconnected", impl);
+		pw_impl_module_schedule_destroy(impl->module);
+		break;
+	case PW_STREAM_STATE_ERROR:
+		pw_log_info("%p: output error: %s", impl, error);
+		break;
 	default:
 		break;
 	}
@@ -523,32 +527,25 @@ static void sink_destroy(void *d)
 	impl->sink = NULL;
 }
 
-static void sink_io_changed(void *data, uint32_t id, void *area, uint32_t size)
-{
-	struct impl *impl = data;
-
-	switch (id) {
-	case SPA_IO_RateMatch:
-		impl->play_rate_match = area;
-		break;
-	}
-}
-
 static void sink_process(void *data)
 {
 	struct impl *impl = data;
 	struct pw_buffer *buf;
 	struct spa_data *d;
-	uint32_t i, index, size;
+	uint32_t i, index, offs, size;
 	int32_t avail;
 
 	if ((buf = pw_stream_dequeue_buffer(impl->sink)) == NULL) {
 		pw_log_debug("out of sink buffers: %m");
 		return;
 	}
+	d = &buf->buffer->datas[0];
+
+	offs = SPA_MIN(d->chunk->offset, d->maxsize);
+	size = SPA_MIN(d->chunk->size, d->maxsize - offs);
 
 	avail = spa_ringbuffer_get_write_index(&impl->play_ring, &index);
-	size = buf->buffer->datas[0].chunk->size;
+
 	if (avail + size > impl->play_ringsize) {
 		uint32_t rindex, drop;
 
@@ -567,7 +564,7 @@ static void sink_process(void *data)
 	}
 
 	if (impl->aec_blocksize == 0) {
-		impl->aec_blocksize = SPA_MAX(size, impl->rec_rate_match->size);
+		impl->aec_blocksize = size;
 		pw_log_debug("Setting AEC block size to %u", impl->aec_blocksize);
 	}
 
@@ -575,12 +572,13 @@ static void sink_process(void *data)
 		/* echo from sink */
 		d = &buf->buffer->datas[i];
 
+		offs = SPA_MIN(d->chunk->offset, d->maxsize);
+		size = SPA_MIN(d->chunk->size, d->maxsize - offs);
+
 		spa_ringbuffer_write_data(&impl->play_ring, impl->play_buffer[i],
 				impl->play_ringsize, index % impl->play_ringsize,
-				SPA_PTROFF(d->data, d->chunk->offset, void),
-				d->chunk->size);
+				SPA_PTROFF(d->data, offs, void), size);
 	}
-
 	spa_ringbuffer_write_update(&impl->play_ring, index + size);
 
 	if (avail + size >= impl->aec_blocksize) {
@@ -608,7 +606,6 @@ static const struct pw_stream_events playback_events = {
 static const struct pw_stream_events sink_events = {
 	PW_VERSION_STREAM_EVENTS,
 	.destroy = sink_destroy,
-	.io_changed = sink_io_changed,
 	.process = sink_process,
 	.state_changed = output_state_changed,
 	.param_changed = output_param_changed
@@ -638,6 +635,10 @@ static int setup_streams(struct impl *impl)
 		pw_properties_set(props, PW_KEY_NODE_LATENCY, str);
 	else if (impl->aec->latency)
 		pw_properties_set(props, PW_KEY_NODE_LATENCY, impl->aec->latency);
+	if ((str = pw_properties_get(impl->source_props, SPA_KEY_AUDIO_CHANNELS)) != NULL)
+		pw_properties_set(props, SPA_KEY_AUDIO_CHANNELS, str);
+	if ((str = pw_properties_get(impl->source_props, SPA_KEY_AUDIO_POSITION)) != NULL)
+		pw_properties_set(props, SPA_KEY_AUDIO_POSITION, str);
 
 	impl->capture = pw_stream_new(impl->core,
 			"Echo-Cancel Capture", props);
@@ -671,6 +672,10 @@ static int setup_streams(struct impl *impl)
 		pw_properties_set(props, PW_KEY_NODE_LATENCY, str);
 	else if (impl->aec->latency)
 		pw_properties_set(props, PW_KEY_NODE_LATENCY, impl->aec->latency);
+	if ((str = pw_properties_get(impl->sink_props, SPA_KEY_AUDIO_CHANNELS)) != NULL)
+		pw_properties_set(props, SPA_KEY_AUDIO_CHANNELS, str);
+	if ((str = pw_properties_get(impl->sink_props, SPA_KEY_AUDIO_POSITION)) != NULL)
+		pw_properties_set(props, SPA_KEY_AUDIO_POSITION, str);
 
 	impl->playback = pw_stream_new(impl->core,
 			"Echo-Cancel Playback", props);
@@ -877,6 +882,7 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	struct pw_properties *props, *aec_props;
 	struct impl *impl;
 	uint32_t id = pw_global_get_id(pw_impl_module_get_global(module));
+	uint32_t pid = getpid();
 	const char *str;
 	const char *path;
 	int res = 0;
@@ -909,14 +915,13 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 		goto error;
 	}
 
-	impl->id = id;
 	impl->module = module;
 	impl->context = context;
 
 	if (pw_properties_get(props, PW_KEY_NODE_GROUP) == NULL)
-		pw_properties_setf(props, PW_KEY_NODE_GROUP, "echo-cancel-%u", id);
+		pw_properties_setf(props, PW_KEY_NODE_GROUP, "echo-cancel-%u-%u", pid, id);
 	if (pw_properties_get(props, PW_KEY_NODE_LINK_GROUP) == NULL)
-		pw_properties_setf(props, PW_KEY_NODE_LINK_GROUP, "echo-cancel-%u", id);
+		pw_properties_setf(props, PW_KEY_NODE_LINK_GROUP, "echo-cancel-%u-%u", pid, id);
 	if (pw_properties_get(props, PW_KEY_NODE_VIRTUAL) == NULL)
 		pw_properties_set(props, PW_KEY_NODE_VIRTUAL, "true");
 
@@ -1052,6 +1057,8 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	copy_props(impl, props, PW_KEY_NODE_LINK_GROUP);
 	copy_props(impl, props, PW_KEY_NODE_VIRTUAL);
 	copy_props(impl, props, PW_KEY_NODE_LATENCY);
+	copy_props(impl, props, SPA_KEY_AUDIO_CHANNELS);
+	copy_props(impl, props, SPA_KEY_AUDIO_POSITION);
 
 	impl->max_buffer_size = pw_properties_get_uint32(props,"buffer.max_size", MAX_BUFSIZE_MS);
 
