@@ -35,7 +35,7 @@
 #include <fcntl.h>
 #include <sys/file.h>
 #include <ctype.h>
-#if HAVE_PWD_H
+#ifdef HAVE_PWD_H
 #include <pwd.h>
 #endif
 #if defined(__FreeBSD__)
@@ -57,6 +57,7 @@
 
 #include "modules/module-protocol-native/connection.h"
 #include "modules/module-protocol-native/defs.h"
+#include "modules/module-protocol-native/protocol-footer.h"
 
 
 #define NAME "protocol-native"
@@ -71,6 +72,54 @@ PW_LOG_TOPIC(mod_topic_connection, "conn." NAME);
 #include <spa/debug/types.h>
 
 /** \page page_module_protocol_native PipeWire Module: Protocol Native
+ *
+ * The native protocol module implements the PipeWire communication between
+ * a client and a server using unix local sockets.
+ *
+ * Normally this module is loaded in both client and server config files
+ * so that they cam communicate.
+ *
+ * ## Module Options
+ *
+ * The module has no options.
+ *
+ * ## General Options
+ *
+ * The name of the core is obtained as:
+ *
+ * - PIPEWIRE_CORE : the environment variable with the name of the core
+ * - \ref PW_KEY_CORE_NAME : in the context properties
+ * - a name based on the process id
+ *
+ * The context will also become a server if:
+ *
+ * - PIPEWIRE_DAEMON : the environment is true
+ * - \ref PW_KEY_CORE_DAEMON : in the context properties is true
+ *
+ * The socket will be located in the directory obtained by looking at the
+ * following environment variables:
+ *
+ * - PIPEWIRE_RUNTIME_DIR
+ * - XDG_RUNTIME_DIR
+ * - USERPROFILE
+ *
+ * When a client connect, the connection will be made to:
+ *
+ * - PIPEWIRE_REMOTE : the environment with the remote name
+ * - \ref PW_KEY_REMOTE_NAME : the property in the context.
+ * - The default remote named "pipewire-0"
+ *
+ * A Special remote named "internal" can be used to make a connection to the
+ * local context. This can be done even when the server is not a daemon. It can
+ * be used to treat a local context as if it was a server.
+ *
+ * ## Example configuration
+ *
+ *\code{.unparsed}
+ * context.modules = [
+   { name = libpipewire-module-protocol-native }
+ * ]
+ *\endcode
  */
 
 #ifndef UNIX_PATH_MAX
@@ -115,6 +164,8 @@ struct client {
 
 	int ref;
 
+	struct footer_core_global_state footer_state;
+
 	unsigned int connected:1;
 	unsigned int disconnecting:1;
 	unsigned int need_flush:1;
@@ -151,6 +202,8 @@ struct client_data {
 	struct pw_protocol_native_connection *connection;
 	struct spa_hook conn_listener;
 
+	struct footer_client_global_state footer_state;
+
 	unsigned int busy:1;
 	unsigned int need_flush:1;
 
@@ -164,7 +217,7 @@ static void debug_msg(const char *prefix, const struct pw_protocol_native_messag
 		      "%s: id:%d op:%d size:%d seq:%d", prefix,
 		      msg->id, msg->opcode, msg->size, msg->seq);
 
-	if ((pod = spa_pod_from_data(msg->data, msg->size, 0, msg->size)) != NULL)
+	if ((pod = get_first_pod_from_data(msg->data, msg->size, 0)) != NULL)
 		spa_debug_pod(0, NULL, pod);
 	else
 		hex = true;
@@ -172,6 +225,49 @@ static void debug_msg(const char *prefix, const struct pw_protocol_native_messag
 		spa_debug_mem(0, msg->data, msg->size);
 }
 
+static void pre_demarshal(struct pw_protocol_native_connection *conn,
+		const struct pw_protocol_native_message *msg,
+		void *object, const struct footer_demarshal *opcodes, size_t n_opcodes)
+{
+	struct spa_pod *footer = NULL;
+	struct spa_pod_parser parser;
+	struct spa_pod_frame f[2];
+	uint32_t opcode;
+	int ret;
+
+	footer = pw_protocol_native_connection_get_footer(conn, msg);
+	if (footer == NULL)
+		return;   /* No valid footer. Ignore silently. */
+
+	/*
+	 * Version 3 footer
+	 *
+	 * spa_pod Struct { [Id opcode, Struct { ... }]* }
+	 */
+
+	spa_pod_parser_pod(&parser, footer);
+	if (spa_pod_parser_push_struct(&parser, &f[0]) < 0) {
+		pw_log_error("malformed message footer");
+		return;
+	}
+
+	while (1) {
+		if (spa_pod_parser_get_id(&parser, &opcode) < 0)
+			break;
+		if (spa_pod_parser_push_struct(&parser, &f[1]) < 0)
+			break;
+		if (opcode < n_opcodes) {
+			if ((ret = opcodes[opcode].demarshal(object, &parser)) < 0)
+				pw_log_error("failed processing message footer (opcode %u): %d (%s)",
+						opcode, ret, spa_strerror(ret));
+		} else {
+			/* Ignore (don't log errors), in case we need to extend this later. */
+			pw_log_debug("unknown message footer opcode %u", opcode);
+		}
+		spa_pod_parser_pop(&parser, &f[1]);
+	}
+}
+
 static int
 process_messages(struct client_data *data)
 {
@@ -213,6 +309,9 @@ process_messages(struct client_data *data)
 		if (debug_messages)
 			debug_msg("<<<<<< in", msg, false);
 
+		pre_demarshal(conn, msg, client, footer_client_demarshal,
+				SPA_N_ELEMENTS(footer_client_demarshal));
+
 		resource = pw_impl_client_find_resource(client, msg->id);
 		if (resource == NULL) {
 			pw_resource_errorf(client->core_resource,
@@ -248,9 +347,12 @@ process_messages(struct client_data *data)
 			continue;
 		}
 
+		resource->refcount++;
 		pw_protocol_native_connection_enter(conn);
 		res = demarshal[msg->opcode].func(resource, msg);
 		pw_protocol_native_connection_leave(conn);
+		pw_resource_unref(resource);
+
 		if (res < 0) {
 			pw_resource_errorf_id(resource, msg->id,
 					res, "invalid message id:%u op:%u (%s)",
@@ -261,6 +363,7 @@ process_messages(struct client_data *data)
 	res = 0;
 done:
 	context->current_client = NULL;
+
 	return res;
 
 error:
@@ -288,14 +391,15 @@ client_busy_changed(void *data, bool busy)
 		pw_loop_signal_event(s->loop, s->resume);
 }
 
-static void handle_client_error(struct pw_impl_client *client, int res)
+static void handle_client_error(struct pw_impl_client *client, int res, const char *msg)
 {
 	if (res == -EPIPE || res == -ECONNRESET)
-		pw_log_info("%p: client %p disconnected", client->protocol, client);
+		pw_log_info("%p: %s: client %p disconnected", client->protocol, msg, client);
 	else
-		pw_log_error("%p: client %p error %d (%s)", client->protocol,
+		pw_log_error("%p: %s: client %p error %d (%s)", client->protocol, msg,
 				client, res, spa_strerror(res));
-	pw_impl_client_destroy(client);
+	if (!client->destroyed)
+		pw_impl_client_destroy(client);
 }
 
 static void
@@ -305,6 +409,8 @@ connection_data(void *data, int fd, uint32_t mask)
 	struct pw_impl_client *client = this->client;
 	int res;
 
+	client->refcount++;
+
 	if (mask & SPA_IO_HUP) {
 		res = -EPIPE;
 		goto error;
@@ -326,9 +432,19 @@ connection_data(void *data, int fd, uint32_t mask)
 		} else if (res != -EAGAIN)
 			goto error;
 	}
+done:
+	pw_impl_client_unref(client);
 	return;
 error:
-	handle_client_error(client, res);
+	handle_client_error(client, res, "connection_data");
+	goto done;
+}
+
+static void client_destroy(void *data)
+{
+	struct client_data *this = data;
+	pw_log_debug("%p: destroy", this);
+	spa_list_remove(&this->protocol_link);
 }
 
 static void client_free(void *data)
@@ -337,8 +453,6 @@ static void client_free(void *data)
 	struct pw_impl_client *client = this->client;
 
 	pw_log_debug("%p: free", this);
-	spa_list_remove(&this->protocol_link);
-
 	spa_hook_remove(&this->client_listener);
 
 	if (this->source)
@@ -351,6 +465,7 @@ static void client_free(void *data)
 
 static const struct pw_impl_client_events client_events = {
 	PW_VERSION_IMPL_CLIENT_EVENTS,
+	.destroy = client_destroy,
 	.free = client_free,
 	.busy_changed = client_busy_changed,
 };
@@ -767,6 +882,9 @@ process_remote(struct client *impl)
 		if (debug_messages)
 			debug_msg("<<<<<< in", msg, false);
 
+		pre_demarshal(conn, msg, this, footer_core_demarshal,
+				SPA_N_ELEMENTS(footer_core_demarshal));
+
 		proxy = pw_core_find_proxy(this, msg->id);
 		if (proxy == NULL || proxy->zombie) {
 			if (proxy == NULL)
@@ -1113,12 +1231,12 @@ static void do_resume(void *_data, uint64_t count)
 	pw_log_debug("flush");
 
 	spa_list_for_each_safe(data, tmp, &this->client_list, protocol_link) {
+		data->client->refcount++;
 		if ((res = process_messages(data)) < 0)
-			goto error;
+			handle_client_error(data->client, res, "do_resume");
+		pw_impl_client_unref(data->client);
 	}
 	return;
-error:
-	handle_client_error(data->client, res);
 }
 
 static const char *
@@ -1225,11 +1343,25 @@ static int impl_ext_get_proxy_fd(struct pw_proxy *proxy, uint32_t index)
 	return pw_protocol_native_connection_get_fd(impl->connection, index);
 }
 
+static void assert_single_pod(struct spa_pod_builder *builder)
+{
+	/*
+	 * Check the invariant that the message we just marshaled
+	 * consists of at most one POD.
+	 */
+	struct spa_pod *pod = builder->data;
+	spa_assert(builder->data == NULL ||
+			builder->state.offset < sizeof(struct spa_pod) ||
+			builder->state.offset == SPA_POD_SIZE(pod));
+}
+
 static int impl_ext_end_proxy(struct pw_proxy *proxy,
 			       struct spa_pod_builder *builder)
 {
 	struct pw_core *core = proxy->core;
 	struct client *impl = SPA_CONTAINER_OF(core->conn, struct client, this);
+	assert_single_pod(builder);
+	marshal_core_footers(&impl->footer_state, core, builder);
 	return core->send_seq = pw_protocol_native_connection_end(impl->connection, builder);
 }
 
@@ -1257,6 +1389,8 @@ static int impl_ext_end_resource(struct pw_resource *resource,
 {
 	struct client_data *data = resource->client->user_data;
 	struct pw_impl_client *client = resource->client;
+	assert_single_pod(builder);
+	marshal_client_footers(&data->footer_state, client, builder);
 	return client->send_seq = pw_protocol_native_connection_end(data->connection, builder);
 }
 static const struct pw_protocol_native_ext protocol_ext_impl = {
