@@ -28,16 +28,19 @@
 #include <spa/debug/buffer.h>
 #include <spa/utils/defs.h>
 #include <spa/utils/string.h>
-#include <pipewire/keys.h>
 #include <pipewire/log.h>
 
 #include "defs.h"
 #include "format.h"
 #include "internal.h"
-#include "media-roles.h"
+#include "log.h"
 #include "message.h"
+#include "remap.h"
 #include "volume.h"
 
+#define MAX_SIZE	(256*1024)
+#define MAX_ALLOCATED	(16*1024 *1024)
+
 #define VOLUME_MUTED ((uint32_t) 0U)
 #define VOLUME_NORM ((uint32_t) 0x10000U)
 #define VOLUME_MAX ((uint32_t) UINT32_MAX/2)
@@ -61,20 +64,6 @@ static inline float volume_to_linear(uint32_t vol)
 	return v * v * v;
 }
 
-static const struct str_map key_table[] = {
-	{ PW_KEY_DEVICE_BUS_PATH, "device.bus_path" },
-	{ PW_KEY_DEVICE_FORM_FACTOR, "device.form_factor" },
-	{ PW_KEY_DEVICE_ICON_NAME, "device.icon_name" },
-	{ PW_KEY_DEVICE_INTENDED_ROLES, "device.intended_roles" },
-	{ PW_KEY_NODE_DESCRIPTION, "device.description" },
-	{ PW_KEY_MEDIA_ICON_NAME, "media.icon_name" },
-	{ PW_KEY_APP_ICON_NAME, "application.icon_name" },
-	{ PW_KEY_APP_PROCESS_MACHINE_ID, "application.process.machine_id" },
-	{ PW_KEY_APP_PROCESS_SESSION_ID, "application.process.session_id" },
-	{ PW_KEY_MEDIA_ROLE, "media.role", media_role_map },
-	{ NULL, NULL },
-};
-
 static int read_u8(struct message *m, uint8_t *val)
 {
 	if (m->offset + 1 > m->length)
@@ -149,7 +138,7 @@ static int read_props(struct message *m, struct pw_properties *props, bool remap
 				TAG_INVALID)) < 0)
 			return res;
 
-		if (remap && (map = str_map_find(key_table, NULL, key)) != NULL) {
+		if (remap && (map = str_map_find(props_key_map, NULL, key)) != NULL) {
 			key = map->pw_str;
 			if (map->child != NULL &&
 			    (map = str_map_find(map->child, NULL, data)) != NULL)
@@ -177,7 +166,10 @@ static int read_arbitrary(struct message *m, const void **val, size_t *length)
 
 static int read_string(struct message *m, char **str)
 {
-	uint32_t n, maxlen = m->length - m->offset;
+	uint32_t n, maxlen;
+	if (m->offset + 1 > m->length)
+		return -ENOSPC;
+	maxlen = m->length - m->offset;
 	n = strnlen(SPA_PTROFF(m->data, m->offset, char), maxlen);
 	if (n == maxlen)
 		return -EINVAL;
@@ -403,8 +395,12 @@ static int ensure_size(struct message *m, uint32_t size)
 
 	alloc = SPA_ROUND_UP_N(SPA_MAX(m->allocated + size, 4096u), 4096u);
 	diff = alloc - m->allocated;
-	if ((data = realloc(m->data, alloc)) == NULL)
+	if ((data = realloc(m->data, alloc)) == NULL) {
+		free(m->data);
+		m->stat->allocated -= m->allocated;
+		m->allocated = 0;
 		return -errno;
+	}
 	m->stat->allocated += diff;
 	m->stat->accumulated += diff;
 	m->data = data;
@@ -560,7 +556,7 @@ static void write_dict(struct message *m, struct spa_dict *dict, bool remap)
 			int l;
 			const struct str_map *map;
 
-			if (remap && (map = str_map_find(key_table, key, NULL)) != NULL) {
+			if (remap && (map = str_map_find(props_key_map, key, NULL)) != NULL) {
 				key = map->pa_str;
 				if (map->child != NULL &&
 				    (map = str_map_find(map->child, val, NULL)) != NULL)
@@ -829,12 +825,12 @@ struct message *message_alloc(struct impl *impl, uint32_t channel, uint32_t size
 	if (!spa_list_is_empty(&impl->free_messages)) {
 		msg = spa_list_first(&impl->free_messages, struct message, link);
 		spa_list_remove(&msg->link);
-		pw_log_trace("using recycled message %p", msg);
+		pw_log_trace("using recycled message %p size:%d", msg, size);
 	} else {
 		if ((msg = calloc(1, sizeof(*msg))) == NULL)
 			return NULL;
 
-		pw_log_trace("new message %p", msg);
+		pw_log_trace("new message %p size:%d", msg, size);
 		msg->stat = &impl->stat;
 		msg->stat->n_allocated++;
 		msg->stat->n_accumulated++;
@@ -858,14 +854,18 @@ void message_free(struct impl *impl, struct message *msg, bool dequeue, bool des
 	if (dequeue)
 		spa_list_remove(&msg->link);
 
+	if (msg->stat->allocated > MAX_ALLOCATED || msg->allocated > MAX_SIZE)
+		destroy = true;
+
 	if (destroy) {
-		pw_log_trace("destroy message %p", msg);
+		pw_log_trace("destroy message %p size:%d", msg, msg->allocated);
 		msg->stat->n_allocated--;
 		msg->stat->allocated -= msg->allocated;
 		free(msg->data);
 		free(msg);
 	} else {
-		pw_log_trace("recycle message %p", msg);
+		pw_log_trace("recycle message %p size:%d/%d", msg, msg->length, msg->allocated);
 		spa_list_append(&impl->free_messages, &msg->link);
+		msg->length = 0;
 	}
 }
