@@ -55,6 +55,62 @@
 #include "module-protocol-pulse/format.h"
 
 /** \page page_module_pulse_tunnel PipeWire Module: Pulse Tunnel
+ *
+ * The pulse-tunnel module provides a source or sink that tunnels all audio to
+ * a remote PulseAudio connection.
+ *
+ * It is usually used with the PulseAudio or module-protocol-pulse on the remote
+ * end to accept the connection.
+ *
+ * This module is usually used together with module-zeroconf-discover that will
+ * automatically load the tunnel with the right parameters based on zeroconf
+ * information.
+ *
+ * ## Module Options
+ *
+ * - `tunnel.mode`: the desired tunnel to create, must be `source` or `sink`.
+ *                  (Default `sink`)
+ * - `pulse.server.address`: the address of the PulseAudio server to tunnel to.
+ * - `pulse.latency`: the latency to end-to-end latency in milliseconds to
+ *                    maintain (Default 200ms).
+ * - `stream.props`: Extra properties for the local stream.
+ *
+ * ## General options
+ *
+ * Options with well-known behavior.
+ *
+ * - \ref PW_KEY_REMOTE_NAME
+ * - \ref PW_KEY_AUDIO_RATE
+ * - \ref PW_KEY_AUDIO_CHANNELS
+ * - \ref SPA_KEY_AUDIO_POSITION
+ * - \ref PW_KEY_NODE_LATENCY
+ * - \ref PW_KEY_NODE_NAME
+ * - \ref PW_KEY_NODE_DESCRIPTION
+ * - \ref PW_KEY_NODE_GROUP
+ * - \ref PW_KEY_NODE_VIRTUAL
+ * - \ref PW_KEY_MEDIA_CLASS
+ * - \ref PW_KEY_NODE_TARGET to specify the remote name or id to link to
+ *
+ * ## Example configuration of a virtual sink
+ *
+ *\code{.unparsed}
+ * context.modules = [
+ * {   name = libpipewire-module-pulse-tunnel
+ *     args = {
+ *         tunnel.mode = sink
+ *         # Set the remote address to tunnel to
+ *         pulse.server.address = "tcp:192.168.1.126"
+ *         #audio.rate=<sample rate>
+ *         #audio.channels=<number of channels>
+ *         #audio.position=<channel map>
+ *         #node.target=<remote target node>
+ *         stream.props = {
+ *             # extra sink properties
+ *         }
+ *     }
+ * }
+ * ]
+ *\endcode
  */
 
 #define NAME "pulse-tunnel"
@@ -72,7 +128,7 @@ PW_LOG_TOPIC_STATIC(mod_topic, "mod." NAME);
 			"[ audio.position=<channel map> ] "			\
 			"pulse.server.address=<address> "			\
 			"pulse.latency=<latency in msec> "			\
-			"[ tunnel.mode=capture|playback "			\
+			"[ tunnel.mode=source|sink "				\
 			"[ stream.props=<properties> ] "
 
 
@@ -91,8 +147,8 @@ static const struct spa_dict_item module_props[] = {
 struct impl {
 	struct pw_context *context;
 
-#define MODE_PLAYBACK	0
-#define MODE_CAPTURE	1
+#define MODE_SINK	0
+#define MODE_SOURCE	1
 	uint32_t mode;
 	struct pw_properties *props;
 
@@ -120,9 +176,12 @@ struct impl {
 	pa_context *pa_context;
 	pa_stream *pa_stream;
 
+	uint32_t target_latency;
+	uint32_t current_latency;
 	uint32_t target_buffer;
 	struct spa_dll dll;
 	float max_error;
+	unsigned resync:1;
 
 	unsigned int do_disconnect:1;
 };
@@ -133,16 +192,21 @@ static void cork_stream(struct impl *impl, bool cork)
 
 	pa_threaded_mainloop_lock(impl->pa_mainloop);
 
-	pw_log_info("corking: %d", cork);
-	if (cork && impl->mode == MODE_PLAYBACK) {
+	pw_log_debug("corking: %d", cork);
+	if (cork && impl->mode == MODE_SINK) {
 		/* When the sink becomes suspended (which is the only case where we
 		 * cork the stream), we don't want to keep any old data around, because
 		 * the old data is most likely unrelated to the audio that will be
 		 * played at the time when the sink starts running again. */
 		if ((operation = pa_stream_flush(impl->pa_stream, NULL, NULL)))
 			pa_operation_unref(operation);
+
 		spa_ringbuffer_init(&impl->ring);
+		memset(impl->buffer, 0, RINGBUFFER_SIZE);
 	}
+	if (!cork)
+		impl->resync = true;
+
 	if ((operation = pa_stream_cork(impl->pa_stream, cork, NULL, NULL)))
 		pa_operation_unref(operation);
 
@@ -182,7 +246,7 @@ static void playback_stream_process(void *d)
 	struct pw_buffer *buf;
 	struct spa_data *bd;
 	int32_t filled;
-	uint32_t write_index, size;
+	uint32_t write_index, offs, size;
 
 	if ((buf = pw_stream_dequeue_buffer(impl->stream)) == NULL) {
 		pw_log_debug("out of buffers: %m");
@@ -190,37 +254,37 @@ static void playback_stream_process(void *d)
 	}
 
 	bd = &buf->buffer->datas[0];
-	size = SPA_MIN(bd->chunk->size, RINGBUFFER_SIZE);
+	offs = SPA_MIN(bd->chunk->offset, bd->maxsize);
+	size = SPA_MIN(bd->chunk->size, bd->maxsize - offs);
+	size = SPA_MIN(size, RINGBUFFER_SIZE);
 
 	filled = spa_ringbuffer_get_write_index(&impl->ring, &write_index);
 
 	if (filled < 0) {
 		pw_log_warn("%p: underrun write:%u filled:%d",
 				impl, write_index, filled);
+	} else if ((uint32_t)filled + size > RINGBUFFER_SIZE) {
+		pw_log_warn("%p: overrun write:%u filled:%d + size:%u > max:%u",
+                                        impl, write_index, filled,
+                                        size, RINGBUFFER_SIZE);
+		impl->resync = true;
 	} else {
 		float error, corr;
 
-		if ((uint32_t)filled + size > impl->target_buffer * 2) {
-			pw_log_warn("%p: overrun write:%u filled:%d size:%u max:%u",
-	                                        impl, write_index, filled,
-	                                        size, RINGBUFFER_SIZE);
-			write_index -= impl->target_buffer;
-			filled -= impl->target_buffer;
-		} else {
-			error = (float)filled - (float)impl->target_buffer;
-			error = SPA_CLAMP(error, -impl->max_error, impl->max_error);
+		error = (float)(impl->current_latency) - (float)impl->target_latency;
+		error = SPA_CLAMP(error, -impl->max_error, impl->max_error);
 
-			pw_log_debug("filled:%u target:%u error:%f corr:%f", filled,
-					impl->target_buffer, error, corr);
-			corr = spa_dll_update(&impl->dll, error);
-			pw_stream_set_control(impl->stream,
-					SPA_PROP_rate, 1, &corr, NULL);
-		}
+		corr = spa_dll_update(&impl->dll, error);
+		pw_log_debug("filled:%u target:%u error:%f corr:%f %u %u", filled,
+				impl->target_buffer, error, corr,
+				impl->current_latency, impl->target_latency);
+		pw_stream_set_control(impl->stream,
+				SPA_PROP_rate, 1, &corr, NULL);
 	}
 	spa_ringbuffer_write_data(&impl->ring,
 				impl->buffer, RINGBUFFER_SIZE,
-                                write_index & RINGBUFFER_MASK,
-                                SPA_PTROFF(bd->data, bd->chunk->offset, void),
+				write_index & RINGBUFFER_MASK,
+				SPA_PTROFF(bd->data, offs, void),
 				size);
 	write_index += size;
 	spa_ringbuffer_write_update(&impl->ring, write_index);
@@ -234,7 +298,7 @@ static void capture_stream_process(void *d)
 	struct pw_buffer *buf;
 	struct spa_data *bd;
 	int32_t avail;
-	uint32_t size, req, read_index;
+	uint32_t size, req, index;
 
 	if ((buf = pw_stream_dequeue_buffer(impl->stream)) == NULL) {
 		pw_log_debug("out of buffers: %m");
@@ -246,42 +310,41 @@ static void capture_stream_process(void *d)
 	if ((req = buf->requested * impl->frame_size) == 0)
 		req = 4096 * impl->frame_size;
 
-	avail = spa_ringbuffer_get_read_index(&impl->ring, &read_index);
-	if (avail <= 0) {
-		size = SPA_MIN(bd->maxsize, req);
+	size = SPA_MIN(bd->maxsize, req);
+
+	avail = spa_ringbuffer_get_read_index(&impl->ring, &index);
+	if (avail < (int32_t)size) {
 		memset(bd->data, 0, size);
 	} else {
 		float error, corr;
 
-		if (avail > (int32_t)impl->target_buffer * 2) {
-			avail -= impl->target_buffer;
-			read_index += impl->target_buffer;
+		if (avail > (int32_t)RINGBUFFER_SIZE) {
+			avail = impl->target_buffer;
+			index += avail - impl->target_buffer;
 		} else {
-			error = (float)impl->target_buffer - (float)avail;
+			error = (float)(impl->current_latency) - (float)impl->target_latency;
 			error = SPA_CLAMP(error, -impl->max_error, impl->max_error);
 
 			corr = spa_dll_update(&impl->dll, error);
 
-			pw_log_debug("avail:%u target:%u error:%f corr:%f", avail,
-					impl->target_buffer, error, corr);
+			pw_log_debug("avail:%u target:%u error:%f corr:%f %u %u", avail,
+					impl->target_buffer, error, corr,
+					impl->current_latency, impl->target_latency);
 			pw_stream_set_control(impl->stream,
 					SPA_PROP_rate, 1, &corr, NULL);
 		}
 
-		size = SPA_MIN(bd->maxsize, (uint32_t)avail);
-		size = SPA_MIN(size, req);
-
 		spa_ringbuffer_read_data(&impl->ring,
 				impl->buffer, RINGBUFFER_SIZE,
-				read_index & RINGBUFFER_MASK,
+				index & RINGBUFFER_MASK,
 				bd->data, size);
 
-		read_index += size;
-		spa_ringbuffer_read_update(&impl->ring, read_index);
-
+		index += size;
+		spa_ringbuffer_read_update(&impl->ring, index);
 	}
 	bd->chunk->offset = 0;
 	bd->chunk->size = size;
+	bd->chunk->stride = impl->frame_size;
 
 	pw_stream_queue_buffer(impl->stream, buf);
 }
@@ -315,7 +378,7 @@ static int create_stream(struct impl *impl)
 	if (impl->stream == NULL)
 		return -errno;
 
-	if (impl->mode == MODE_CAPTURE) {
+	if (impl->mode == MODE_SOURCE) {
 		pw_stream_add_listener(impl->stream,
 				&impl->stream_listener,
 				&capture_stream_events, impl);
@@ -331,14 +394,14 @@ static int create_stream(struct impl *impl)
 			SPA_PARAM_EnumFormat, &impl->info);
 
 	spa_zero(latency);
-	latency.direction = impl->mode == MODE_CAPTURE ? PW_DIRECTION_OUTPUT : PW_DIRECTION_INPUT;
+	latency.direction = impl->mode == MODE_SOURCE ? PW_DIRECTION_OUTPUT : PW_DIRECTION_INPUT;
 	latency.min_ns = latency.max_ns = impl->latency_msec * SPA_NSEC_PER_MSEC;
 
 	params[n_params++] = spa_latency_build(&b,
 			SPA_PARAM_Latency, &latency);
 
 	if ((res = pw_stream_connect(impl->stream,
-			impl->mode == MODE_CAPTURE ? PW_DIRECTION_OUTPUT : PW_DIRECTION_INPUT,
+			impl->mode == MODE_SOURCE ? PW_DIRECTION_OUTPUT : PW_DIRECTION_INPUT,
 			PW_ID_ANY,
 			PW_STREAM_FLAG_AUTOCONNECT |
 			PW_STREAM_FLAG_MAP_BUFFERS |
@@ -385,16 +448,18 @@ static void stream_read_request_cb(pa_stream *s, size_t length, void *userdata)
 {
 	struct impl *impl = userdata;
 	int32_t filled;
-	uint32_t write_index;
+	uint32_t index;
+	pa_usec_t latency;
+	int negative;
 
-	filled = spa_ringbuffer_get_write_index(&impl->ring, &write_index);
+	filled = spa_ringbuffer_get_write_index(&impl->ring, &index);
 
 	if (filled < 0) {
 		pw_log_warn("%p: underrun write:%u filled:%d",
-				impl, write_index, filled);
+				impl, index, filled);
 	} else if (filled + length > RINGBUFFER_SIZE) {
 		pw_log_warn("%p: overrun write:%u filled:%d",
-				impl, write_index, filled);
+				impl, index, filled);
 	}
 	while (length > 0) {
 		const void *p;
@@ -415,26 +480,45 @@ static void stream_read_request_cb(pa_stream *s, size_t length, void *userdata)
 
 			spa_ringbuffer_write_data(&impl->ring,
 					impl->buffer, RINGBUFFER_SIZE,
-					write_index & RINGBUFFER_MASK,
+					index & RINGBUFFER_MASK,
 					p ? p : impl->empty, to_write);
 
-			write_index += to_write;
+			index += to_write;
 			p = p ? SPA_PTROFF(p, to_write, void) : NULL;
 			nbytes -= to_write;
 			length -= to_write;
+			filled += to_write;
 		}
 		pa_stream_drop(impl->pa_stream);
 	}
-	spa_ringbuffer_write_update(&impl->ring, write_index);
+
+	pa_stream_get_latency(impl->pa_stream, &latency, &negative);
+	impl->current_latency = latency * impl->info.rate / SPA_USEC_PER_SEC;
+	impl->current_latency += filled / impl->frame_size;
+
+	spa_ringbuffer_write_update(&impl->ring, index);
 }
 
 static void stream_write_request_cb(pa_stream *s, size_t length, void *userdata)
 {
 	struct impl *impl = userdata;
 	int32_t avail;
-	uint32_t read_index, len, offset, l0, l1;
+	uint32_t index, len, offset, l0, l1;
+	pa_usec_t latency;
+	int negative;
 
-	avail = spa_ringbuffer_get_read_index(&impl->ring, &read_index);
+	if (impl->resync) {
+		impl->resync = false;
+		avail = length + impl->target_buffer;
+		spa_ringbuffer_get_write_index(&impl->ring, &index);
+		index -= avail;
+	} else {
+		avail = spa_ringbuffer_get_read_index(&impl->ring, &index);
+	}
+
+	pa_stream_get_latency(impl->pa_stream, &latency, &negative);
+	impl->current_latency = latency * impl->info.rate / SPA_USEC_PER_SEC;
+	impl->current_latency += avail / impl->frame_size;
 
 	while (avail < (int32_t)length) {
 		/* send silence for the data we don't have */
@@ -447,7 +531,7 @@ static void stream_write_request_cb(pa_stream *s, size_t length, void *userdata)
 	if (length > 0 && avail >= (int32_t)length) {
 		/* always send as much as is requested */
 		len = length;
-		offset = read_index & RINGBUFFER_MASK;
+		offset = index & RINGBUFFER_MASK;
 		l0 = SPA_MIN(len, RINGBUFFER_SIZE - offset);
 		l1 = len - l0;
 
@@ -460,10 +544,22 @@ static void stream_write_request_cb(pa_stream *s, size_t length, void *userdata)
 					impl->buffer, l1,
 					NULL, 0, PA_SEEK_RELATIVE);
 		}
-		read_index += len;
-		spa_ringbuffer_read_update(&impl->ring, read_index);
+		index += len;
+		spa_ringbuffer_read_update(&impl->ring, index);
 	}
 }
+static void stream_underflow_cb(pa_stream *s, void *userdata)
+{
+	struct impl *impl = userdata;
+	pw_log_warn("underflow");
+	impl->resync = true;
+}
+static void stream_overflow_cb(pa_stream *s, void *userdata)
+{
+	struct impl *impl = userdata;
+	pw_log_warn("overflow");
+	impl->resync = true;
+}
 
 static void stream_latency_update_cb(pa_stream *s, void *userdata)
 {
@@ -553,6 +649,8 @@ static int create_pulse_stream(struct impl *impl)
 	pa_stream_set_state_callback(impl->pa_stream, stream_state_cb, impl);
 	pa_stream_set_read_callback(impl->pa_stream, stream_read_request_cb, impl);
 	pa_stream_set_write_callback(impl->pa_stream, stream_write_request_cb, impl);
+	pa_stream_set_underflow_callback(impl->pa_stream, stream_underflow_cb, impl);
+	pa_stream_set_overflow_callback(impl->pa_stream, stream_overflow_cb, impl);
 	pa_stream_set_latency_update_callback(impl->pa_stream, stream_latency_update_cb, impl);
 
 	remote_node_target = pw_properties_get(impl->props, PW_KEY_NODE_TARGET);
@@ -564,10 +662,12 @@ static int create_pulse_stream(struct impl *impl)
 
 	latency_bytes = pa_usec_to_bytes(impl->latency_msec * SPA_USEC_PER_MSEC, &ss);
 
+	impl->target_latency = latency_bytes / impl->frame_size;
+
 	/* half in our buffer, half in the network + remote */
 	impl->target_buffer = latency_bytes / 2;
 
-	if (impl->mode == MODE_CAPTURE) {
+	if (impl->mode == MODE_SOURCE) {
 		bufferattr.fragsize = latency_bytes / 2;
 
 		res = pa_stream_connect_record(impl->pa_stream,
@@ -833,10 +933,10 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	spa_dll_init(&impl->dll);
 
 	if ((str = pw_properties_get(props, "tunnel.mode")) != NULL) {
-		if (spa_streq(str, "capture")) {
-			impl->mode = MODE_CAPTURE;
-		} else if (spa_streq(str, "playback")) {
-			impl->mode = MODE_PLAYBACK;
+		if (spa_streq(str, "source")) {
+			impl->mode = MODE_SOURCE;
+		} else if (spa_streq(str, "sink")) {
+			impl->mode = MODE_SINK;
 		} else {
 			pw_log_error("invalid tunnel.mode '%s'", str);
 			res = -EINVAL;
@@ -847,8 +947,6 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	impl->latency_msec = pw_properties_get_uint32(props, "pulse.latency", DEFAULT_LATENCY_MSEC);
 
 
-	if (pw_properties_get(props, PW_KEY_NODE_WANT_DRIVER) == NULL)
-		pw_properties_set(props, PW_KEY_NODE_WANT_DRIVER, "true");
 	if (pw_properties_get(props, PW_KEY_NODE_VIRTUAL) == NULL)
 		pw_properties_set(props, PW_KEY_NODE_VIRTUAL, "true");
 	if (pw_properties_get(props, PW_KEY_NODE_NETWORK) == NULL)
@@ -856,7 +954,7 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 
 	if (pw_properties_get(props, PW_KEY_MEDIA_CLASS) == NULL)
 		pw_properties_set(props, PW_KEY_MEDIA_CLASS,
-				impl->mode == MODE_PLAYBACK ?
+				impl->mode == MODE_SINK ?
 					"Audio/Sink" : "Audio/Source");
 
 	if ((str = pw_properties_get(props, "stream.props")) != NULL)
@@ -869,7 +967,6 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 	copy_props(impl, props, PW_KEY_NODE_NAME);
 	copy_props(impl, props, PW_KEY_NODE_DESCRIPTION);
 	copy_props(impl, props, PW_KEY_NODE_GROUP);
-	copy_props(impl, props, PW_KEY_NODE_WANT_DRIVER);
 	copy_props(impl, props, PW_KEY_NODE_LATENCY);
 	copy_props(impl, props, PW_KEY_NODE_VIRTUAL);
 	copy_props(impl, props, PW_KEY_MEDIA_CLASS);
@@ -884,7 +981,7 @@ int pipewire__module_init(struct pw_impl_module *module, const char *args)
 		goto error;
 	}
 	spa_dll_set_bw(&impl->dll, SPA_DLL_BW_MIN, 128, impl->info.rate);
-	impl->max_error = 256.0f * impl->frame_size;
+	impl->max_error = 256.0f;
 
 	impl->core = pw_context_get_object(impl->context, PW_TYPE_INTERFACE_Core);
 	if (impl->core == NULL) {
