@@ -89,7 +89,7 @@ struct support {
 	struct spa_interface i18n_iface;
 	struct spa_support support[MAX_SUPPORT];
 	uint32_t n_support;
-	unsigned int initialized:1;
+	uint32_t init_count;
 	unsigned int in_valgrind:1;
 	unsigned int no_color:1;
 	unsigned int no_config:1;
@@ -252,6 +252,10 @@ static struct spa_handle *load_spa_handle(const char *lib,
 	plugin = NULL;
 	res = -ENOENT;
 
+	if (sup->plugin_dir == NULL) {
+		pw_log_error("load lib: plugin directory undefined, set SPA_PLUGIN_DIR");
+		goto error_out;
+	}
 	while ((p = pw_split_walk(sup->plugin_dir, ":", &len, &state))) {
 		if ((plugin = open_plugin(&sup->registry, p, len, lib)) != NULL)
 			break;
@@ -409,11 +413,6 @@ static const char *i18n_ntext(void *object, const char *msgid, const char *msgid
 
 static void init_i18n(struct support *support)
 {
-	/* Load locale from the environment. */
-	setlocale(LC_ALL, "");
-	/* Set LC_NUMERIC to C so that floating point strings are consistently
-	 * formatted and parsed across locales. */
-	setlocale(LC_NUMERIC, "C");
 	bindtextdomain(GETTEXT_PACKAGE, LOCALEDIR);
 	bind_textdomain_codeset(GETTEXT_PACKAGE, "UTF-8");
 	pw_set_domain(GETTEXT_PACKAGE);
@@ -575,8 +574,7 @@ out:
  * Initialize the PipeWire system, parse and modify any parameters given
  * by \a argc and \a argv and set up debugging.
  *
- * The environment variable \a PIPEWIRE_DEBUG
- *
+ * This function can be called multiple times.
  */
 SPA_EXPORT
 void pw_init(int *argc, char **argv[])
@@ -590,7 +588,7 @@ void pw_init(int *argc, char **argv[])
 	char level[32];
 
 	pthread_mutex_lock(&init_lock);
-	if (support->initialized)
+	if (support->init_count > 0)
 		goto done;
 
 	pthread_mutex_lock(&support_lock);
@@ -662,12 +660,25 @@ void pw_init(int *argc, char **argv[])
 	add_i18n(support);
 
 	pw_log_info("version %s", pw_get_library_version());
-	support->initialized = true;
 	pthread_mutex_unlock(&support_lock);
 done:
+	support->init_count++;
 	pthread_mutex_unlock(&init_lock);
 }
 
+/** Deinitialize PipeWire
+ *
+ * Deinitialize the PipeWire system and free up all resources allocated
+ * by pw_init().
+ *
+ * Before 0.3.49 this function can only be called once after which the pipewire
+ * library can not be used again. This is usually called by test programs to
+ * check for memory leaks.
+ *
+ * Since 0.3.49 this function must be paired with an equal amount of pw_init()
+ * calls to deinitialize the PipeWire library. The PipeWire library can be
+ * used again after being deinitialized with a new pw_init() call.
+ */
 SPA_EXPORT
 void pw_deinit(void)
 {
@@ -676,6 +687,11 @@ void pw_deinit(void)
 	struct plugin *p;
 
 	pthread_mutex_lock(&init_lock);
+	if (support->init_count == 0)
+		goto done;
+	if (--support->init_count > 0)
+		goto done;
+
 	pthread_mutex_lock(&support_lock);
 	pw_log_set(NULL);
 	spa_list_consume(p, &registry->plugins, link) {
@@ -689,6 +705,7 @@ void pw_deinit(void)
 	free(support->i18n_domain);
 	spa_zero(global_support);
 	pthread_mutex_unlock(&support_lock);
+done:
 	pthread_mutex_unlock(&init_lock);
 
 }
@@ -700,7 +717,6 @@ void pw_deinit(void)
  *
  * Debugging categories can be enabled by using the PIPEWIRE_DEBUG
  * environment variable
- *
  */
 SPA_EXPORT
 bool pw_debug_is_category_enabled(const char *name)
