@@ -88,18 +88,6 @@ struct data;
 
 typedef int (*fill_fn)(struct data *d, void *dest, unsigned int n_frames);
 
-struct target {
-	struct spa_list link;
-	uint32_t id;
-#define TARGET_TYPE_SINK	0
-#define TARGET_TYPE_SOURCE	1
-#define TARGET_TYPE_STREAM	2
-	uint32_t type;
-	char *name;
-	char *desc;
-	int prio;
-};
-
 struct channelmap {
 	int n_channels;
 	int channels[SPA_AUDIO_MAX_CHANNELS];
@@ -110,16 +98,12 @@ struct data {
 	struct pw_context *context;
 	struct pw_core *core;
 	struct spa_hook core_listener;
-	struct pw_registry *registry;
-	struct spa_hook registry_listener;
-	struct pw_metadata *metadata;
-	struct spa_hook metadata_listener;
-	char default_sink[1024];
-	char default_source[1024];
 
 	struct pw_stream *stream;
 	struct spa_hook stream_listener;
 
+	struct spa_source *timer;
+
 	enum mode mode;
 	bool verbose;
 #define TYPE_PCM	0
@@ -155,14 +139,7 @@ struct data {
 
 	fill_fn fill;
 
-	uint32_t target_id;
-	bool list_targets;
-	bool targets_listed;
-	struct spa_list targets;
-	int sync;
-
 	struct spa_io_position *position;
-	struct spa_io_rate_match *rate_match;
 	bool drained;
 	uint64_t clock_time;
 
@@ -611,39 +588,6 @@ static void channelmap_print(struct channelmap *map)
 	}
 }
 
-static void
-target_destroy(struct target *target)
-{
-	if (!target)
-		return;
-	if (target->name)
-		free(target->name);
-	if (target->desc)
-		free(target->desc);
-	free(target);
-}
-
-static struct target *
-target_create(uint32_t id, uint32_t type, const char *name, const char *desc, int prio)
-{
-	struct target *target;
-
-	target = malloc(sizeof(*target));
-	if (!target)
-		return NULL;
-	target->id = id;
-	target->type = type;
-	target->name = strdup(name);
-	target->desc = strdup(desc ? : "");
-	target->prio = prio;
-
-	if (!target->name || !target->desc) {
-		target_destroy(target);
-		return NULL;
-	}
-	return target;
-}
-
 static void on_core_info(void *userdata, const struct pw_core_info *info)
 {
 	struct data *data = userdata;
@@ -653,20 +597,6 @@ static void on_core_info(void *userdata, const struct pw_core_info *info)
 				info->id, info->name);
 }
 
-static void on_core_done(void *userdata, uint32_t id, int seq)
-{
-	struct data *data = userdata;
-
-	if (data->verbose)
-		printf("core done\n");
-
-	/* if we're listing targets just exist */
-	if (data->sync == seq && data->list_targets) {
-		data->targets_listed = true;
-		pw_main_loop_quit(data->loop);
-	}
-}
-
 static void on_core_error(void *userdata, uint32_t id, int seq, int res, const char *message)
 {
 	struct data *data = userdata;
@@ -681,158 +611,9 @@ static void on_core_error(void *userdata, uint32_t id, int seq, int res, const c
 static const struct pw_core_events core_events = {
 	PW_VERSION_CORE_EVENTS,
 	.info = on_core_info,
-	.done = on_core_done,
 	.error = on_core_error,
 };
 
-static int json_object_find(const char *obj, const char *key, char *value, size_t len)
-{
-	struct spa_json it[2];
-	const char *v;
-	char k[128];
-
-	spa_json_init(&it[0], obj, strlen(obj));
-	if (spa_json_enter_object(&it[0], &it[1]) <= 0)
-		return -EINVAL;
-
-	while (spa_json_get_string(&it[1], k, sizeof(k)) > 0) {
-		if (spa_streq(k, key)) {
-			if (spa_json_get_string(&it[1], value, len) <= 0)
-				continue;
-			return 0;
-		} else {
-			if (spa_json_next(&it[1], &v) <= 0)
-				break;
-		}
-	}
-	return -ENOENT;
-}
-
-static int metadata_property(void *object,
-		uint32_t subject, const char *key, const char *type, const char *value)
-{
-	struct data *data = object;
-
-	if (subject == PW_ID_CORE) {
-		if (key == NULL || spa_streq(key, "default.audio.sink")) {
-			if (value == NULL ||
-			    json_object_find(value, "name",
-					data->default_sink, sizeof(data->default_sink)) < 0)
-				data->default_sink[0] = '\0';
-		}
-		if (key == NULL || spa_streq(key, "default.audio.source")) {
-			if (value == NULL ||
-			    json_object_find(value, "name",
-					data->default_source, sizeof(data->default_source)) < 0)
-				data->default_source[0] = '\0';
-		}
-	}
-	return 0;
-}
-
-static const struct pw_metadata_events metadata_events = {
-	PW_VERSION_METADATA_EVENTS,
-	.property = metadata_property,
-};
-
-static void registry_event_global(void *userdata, uint32_t id,
-		uint32_t permissions, const char *type, uint32_t version,
-		const struct spa_dict *props)
-{
-	struct data *data = userdata;
-	const struct spa_dict_item *item;
-	const char *name, *desc, *media_class, *prio_session;
-	int prio;
-	struct target *target;
-	uint32_t ttype;
-
-	/* only once */
-	if (data->targets_listed)
-		return;
-
-	/* must be listing targets and interface must be a node */
-	if (!data->list_targets)
-		return;
-
-	if (spa_streq(type, PW_TYPE_INTERFACE_Metadata)) {
-		if (data->metadata != NULL)
-			return;
-		if ((name = spa_dict_lookup(props, PW_KEY_METADATA_NAME)) != NULL &&
-		    !spa_streq(name, "default"))
-			return;
-
-		data->metadata = pw_registry_bind(data->registry,
-				id, type, PW_VERSION_METADATA, 0);
-		pw_metadata_add_listener(data->metadata,
-				&data->metadata_listener,
-				&metadata_events, data);
-
-		data->sync = pw_core_sync(data->core, 0, data->sync);
-
-	} else if (spa_streq(type, PW_TYPE_INTERFACE_Node)) {
-		name = spa_dict_lookup(props, PW_KEY_NODE_NAME);
-		desc = spa_dict_lookup(props, PW_KEY_NODE_DESCRIPTION);
-		media_class = spa_dict_lookup(props, PW_KEY_MEDIA_CLASS);
-		prio_session = spa_dict_lookup(props, PW_KEY_PRIORITY_SESSION);
-
-		/* name and media class must exist */
-		if (!name || !media_class)
-			return;
-
-		if (desc == NULL)
-			desc = name;
-
-		/* get allowed mode from the media class */
-		if (spa_streq(media_class, "Audio/Source")) {
-			if (data->mode != mode_record)
-				return;
-			ttype = TARGET_TYPE_SOURCE;
-		}
-		else if (spa_streq(media_class, "Stream/Output/Audio")) {
-			if (data->mode != mode_record)
-				return;
-			ttype = TARGET_TYPE_STREAM;
-		}
-		else if (spa_streq(media_class, "Audio/Sink")) {
-			if (data->mode != mode_playback &&
-			    data->mode != mode_record)
-				return;
-			ttype = TARGET_TYPE_SINK;
-		}
-		else
-			return;
-
-		prio = prio_session ? atoi(prio_session) : -1;
-
-		if (data->verbose) {
-			printf("registry: id=%"PRIu32" type=%s name=\"%s\" media_class=\"%s\" desc=\"%s\" prio=%d\n",
-					id, type, name, media_class, desc ? : "", prio);
-
-			spa_dict_for_each(item, props) {
-				printf("\t\t%s = \"%s\"\n", item->key, item->value);
-			}
-		}
-
-		target = target_create(id, ttype, name, desc, prio);
-		if (target)
-			spa_list_append(&data->targets, &target->link);
-	}
-}
-
-static void registry_event_global_remove(void *userdata, uint32_t id)
-{
-	struct data *data = userdata;
-
-	if (data->verbose)
-		printf("registry: remove id=%"PRIu32"\n", id);
-}
-
-static const struct pw_registry_events registry_events = {
-	PW_VERSION_REGISTRY_EVENTS,
-	.global = registry_event_global,
-	.global_remove = registry_event_global_remove,
-};
-
 static void
 on_state_changed(void *userdata, enum pw_stream_state old,
 		 enum pw_stream_state state, const char *error)
@@ -845,29 +626,41 @@ on_state_changed(void *userdata, enum pw_stream_state old,
 				pw_stream_state_as_string(old),
 				pw_stream_state_as_string(state));
 
-	if (state == PW_STREAM_STATE_STREAMING && !data->volume_is_set) {
+	switch (state) {
+	case PW_STREAM_STATE_STREAMING:
+		if (!data->volume_is_set) {
+			ret = pw_stream_set_control(data->stream,
+					SPA_PROP_volume, 1, &data->volume,
+					0);
+			if (data->verbose)
+				printf("stream set volume to %.3f - %s\n", data->volume,
+						ret == 0 ? "success" : "FAILED");
 
-		ret = pw_stream_set_control(data->stream,
-				SPA_PROP_volume, 1, &data->volume,
-				0);
-		if (data->verbose)
-			printf("set stream volume to %.3f - %s\n", data->volume,
-					ret == 0 ? "success" : "FAILED");
-
-		data->volume_is_set = true;
-
-	}
-
-	if (state == PW_STREAM_STATE_STREAMING) {
-		if (data->verbose)
+			data->volume_is_set = true;
+		}
+		if (data->verbose) {
+			struct timespec timeout = {0, 1}, interval = {1, 0};
+			struct pw_loop *l = pw_main_loop_get_loop(data->loop);
+			pw_loop_update_timer(l, data->timer, &timeout, &interval, false);
 			printf("stream node %"PRIu32"\n",
 				pw_stream_get_node_id(data->stream));
-	}
-	if (state == PW_STREAM_STATE_ERROR) {
+		}
+		break;
+	case PW_STREAM_STATE_PAUSED:
+		if (data->verbose) {
+			struct timespec timeout = {0, 0}, interval = {0, 0};
+			struct pw_loop *l = pw_main_loop_get_loop(data->loop);
+			pw_loop_update_timer(l, data->timer, &timeout, &interval, false);
+		}
+		break;
+	case PW_STREAM_STATE_ERROR:
 		printf("stream node %"PRIu32" error: %s\n",
 				pw_stream_get_node_id(data->stream),
 				error);
 		pw_main_loop_quit(data->loop);
+		break;
+	default:
+		break;
 	}
 }
 
@@ -880,10 +673,6 @@ on_io_changed(void *userdata, uint32_t id, void *data, uint32_t size)
 	case SPA_IO_Position:
 		d->position = data;
 		break;
-	case SPA_IO_RateMatch:
-		if (d->data_type == TYPE_PCM)
-			d->rate_match = data;
-		break;
 	default:
 		break;
 	}
@@ -897,7 +686,8 @@ on_param_changed(void *userdata, uint32_t id, const struct spa_pod *param)
 	int err;
 
 	if (data->verbose)
-		printf("stream param change: id=%"PRIu32"\n", id);
+		printf("stream param change: %s\n",
+			spa_debug_type_find_name(spa_type_param, id));
 
 	if (id != SPA_PARAM_Format || param == NULL)
 		return;
@@ -919,10 +709,11 @@ on_param_changed(void *userdata, uint32_t id, const struct spa_pod *param)
 	data->stride = data->dsf.layout.channels * SPA_ABS(data->dsf.layout.interleave);
 
 	if (data->verbose) {
-		printf("DSD out: channels:%d bitorder:%s interleave:%d\n",
+		printf("DSD: channels:%d bitorder:%s interleave:%d stride:%d\n",
 				data->dsf.layout.channels,
 				data->dsf.layout.lsb ? "lsb" : "msb",
-				data->dsf.layout.interleave);
+				data->dsf.layout.interleave,
+				data->stride);
 	}
 }
 
@@ -951,18 +742,22 @@ static void on_process(void *userdata)
 	if (data->mode == mode_playback) {
 
 		n_frames = d->maxsize / data->stride;
-		if (data->rate_match && data->rate_match->size > 0)
-			n_frames = SPA_MIN((uint32_t)n_frames, data->rate_match->size);
+		n_frames = SPA_MIN(n_frames, (int)b->requested);
 
 		n_fill_frames = data->fill(data, p, n_frames);
 
-		if (n_fill_frames > 0) {
+		if (n_fill_frames > 0 || n_frames == 0) {
 			d->chunk->offset = 0;
 			d->chunk->stride = data->stride;
 			d->chunk->size = n_fill_frames * data->stride;
 			have_data = true;
-		} else if (n_fill_frames < 0)
+			b->size = n_frames;
+		} else if (n_fill_frames < 0) {
 			fprintf(stderr, "fill error %d\n", n_fill_frames);
+		} else {
+			if (data->verbose)
+				printf("drain start\n");
+		}
 	} else {
 		offset = SPA_MIN(d->chunk->offset, d->maxsize);
 		size = SPA_MIN(d->chunk->size, d->maxsize - offset);
@@ -1015,11 +810,14 @@ static void do_print_delay(void *userdata, uint64_t expirations)
 {
 	struct data *data = userdata;
 	struct pw_time time;
-	pw_stream_get_time(data->stream, &time);
-	printf("now=%"PRIi64" rate=%u/%u ticks=%"PRIu64" delay=%"PRIi64" queued=%"PRIu64"\n",
+	pw_stream_get_time_n(data->stream, &time, sizeof(time));
+	printf("stream time: now:%"PRIi64" rate:%u/%u ticks:%"PRIu64
+			" delay:%"PRIi64" queued:%"PRIu64
+			" buffered:%"PRIi64" buffers:%u avail:%u\n",
 		time.now,
 		time.rate.num, time.rate.denom,
-		time.ticks, time.delay, time.queued);
+		time.ticks, time.delay, time.queued, time.buffered,
+		time.queued_buffers, time.avail_buffers);
 }
 
 enum {
@@ -1034,7 +832,6 @@ enum {
 	OPT_CHANNELMAP,
 	OPT_FORMAT,
 	OPT_VOLUME,
-	OPT_LIST_TARGETS,
 };
 
 static const struct option long_options[] = {
@@ -1053,6 +850,7 @@ static const struct option long_options[] = {
 	{ "media-role",		required_argument, NULL, OPT_MEDIA_ROLE },
 	{ "target",		required_argument, NULL, OPT_TARGET },
 	{ "latency",		required_argument, NULL, OPT_LATENCY },
+	{ "properties",		required_argument, NULL, 'P' },
 
 	{ "rate",		required_argument, NULL, OPT_RATE },
 	{ "channels",		required_argument, NULL, OPT_CHANNELS },
@@ -1061,8 +859,6 @@ static const struct option long_options[] = {
 	{ "volume",		required_argument, NULL, OPT_VOLUME },
 	{ "quality",		required_argument, NULL, 'q' },
 
-	{ "list-targets",	no_argument, NULL, OPT_LIST_TARGETS },
-
 	{ NULL, 0, NULL, 0 }
 };
 
@@ -1073,7 +869,7 @@ static void show_usage(const char *name, bool is_error)
 	fp = is_error ? stderr : stdout;
 
         fprintf(fp,
-	   _("%s [options] <file>\n"
+	   _("%s [options] [<file>|-]\n"
              "  -h, --help                            Show this help\n"
              "      --version                         Show version\n"
              "  -v, --verbose                         Enable verbose operations\n"
@@ -1090,7 +886,7 @@ static void show_usage(const char *name, bool is_error)
 	     "                                          Xunit (unit = s, ms, us, ns)\n"
 	     "                                          or direct samples (256)\n"
 	     "                                          the rate is the one of the source file\n"
-	     "      --list-targets                    List available targets for --target\n"
+	     "  -P  --properties                      Set node properties\n"
 	     "\n"),
 	     DEFAULT_MEDIA_TYPE,
 	     DEFAULT_MEDIA_CATEGORY_PLAYBACK,
@@ -1222,12 +1018,12 @@ static int setup_midifile(struct data *data)
 			data->mode == mode_playback ? "r" : "w",
 			&data->midi.info);
 	if (data->midi.file == NULL) {
-		fprintf(stderr, "error: can't read midi file '%s': %m\n", data->filename);
+		fprintf(stderr, "midifile: can't read midi file '%s': %m\n", data->filename);
 		return -errno;
 	}
 
 	if (data->verbose)
-		printf("opened file \"%s\" format %08x ntracks:%d div:%d\n",
+		printf("midifile: opened file \"%s\" format %08x ntracks:%d div:%d\n",
 				data->filename,
 				data->midi.info.format, data->midi.info.ntracks,
 				data->midi.info.division);
@@ -1264,12 +1060,12 @@ static int setup_dsffile(struct data *data)
 
 	data->dsf.file = dsf_file_open(data->filename, "r", &data->dsf.info);
 	if (data->dsf.file == NULL) {
-		fprintf(stderr, "error: can't read dsf file '%s': %m\n", data->filename);
+		fprintf(stderr, "dsffile: can't read dsf file '%s': %m\n", data->filename);
 		return -errno;
 	}
 
 	if (data->verbose)
-		printf("opened file \"%s\" channels:%d rate:%d samples:%"PRIu64" bitorder:%s\n",
+		printf("dsffile: opened file \"%s\" channels:%d rate:%d samples:%"PRIu64" bitorder:%s\n",
 				data->filename,
 				data->dsf.info.channels, data->dsf.info.rate,
 				data->dsf.info.samples,
@@ -1280,6 +1076,62 @@ static int setup_dsffile(struct data *data)
 	return 0;
 }
 
+struct format_info {
+	const char *name;
+	uint32_t spa_format;
+	uint32_t width;
+} format_info[] = {
+	{  "s8", SPA_AUDIO_FORMAT_S8, 1 },
+	{  "u8", SPA_AUDIO_FORMAT_U8, 1 },
+	{  "s16", SPA_AUDIO_FORMAT_S16, 2 },
+	{  "s24", SPA_AUDIO_FORMAT_S24, 3 },
+	{  "s32", SPA_AUDIO_FORMAT_S32, 4 },
+	{  "f32", SPA_AUDIO_FORMAT_F32, 4 },
+	{  "f64", SPA_AUDIO_FORMAT_F32, 8 },
+};
+
+static struct format_info *format_info_by_name(const char *str)
+{
+	uint32_t i;
+	for (i = 0; i < SPA_N_ELEMENTS(format_info); i++)
+		if (spa_streq(str, format_info[i].name))
+			return &format_info[i];
+	return NULL;
+}
+
+static int stdout_record(struct data *d, void *src, unsigned int n_frames)
+{
+	return fwrite(src, d->stride, n_frames, stdout);
+}
+
+static int stdin_play(struct data *d, void *src, unsigned int n_frames)
+{
+	return fread(src, d->stride, n_frames, stdin);
+}
+
+static int setup_pipe(struct data *data)
+{
+	struct format_info *info;
+
+	if (data->format == NULL)
+		data->format = DEFAULT_FORMAT;
+	if (data->channels == 0)
+		data->channels = DEFAULT_CHANNELS;
+	if (data->rate == 0)
+		data->rate = DEFAULT_RATE;
+	if (data->channelmap.n_channels == 0)
+		channelmap_default(&data->channelmap, data->channels);
+
+	info = format_info_by_name(data->format);
+	if (info == NULL)
+		return -EINVAL;
+
+	data->spa_format = info->spa_format;
+	data->stride = info->width * data->channels;
+	data->fill = data->mode == mode_playback ?  stdin_play : stdout_record;
+	return 0;
+}
+
 static int fill_properties(struct data *data)
 {
 	static const char * const table[] = {
@@ -1327,6 +1179,45 @@ static int fill_properties(struct data *data)
 
 	return 0;
 }
+static void format_from_filename(SF_INFO *info, const char *filename)
+{
+	int i, count = 0;
+	int format = -1;
+
+#if __BYTE_ORDER == __BIG_ENDIAN
+	info->format |= SF_ENDIAN_BIG;
+#else
+	info->format |= SF_ENDIAN_LITTLE;
+#endif
+
+	if (sf_command(NULL, SFC_GET_FORMAT_MAJOR_COUNT, &count, sizeof(int)) != 0)
+		count = 0;
+
+	for (i = 0; i < count; i++) {
+		SF_FORMAT_INFO fi;
+
+		spa_zero(fi);
+		fi.format = i;
+		if (sf_command(NULL, SFC_GET_FORMAT_MAJOR, &fi, sizeof(fi)) != 0)
+			continue;
+
+		if (spa_strendswith(filename, fi.extension)) {
+			format = fi.format;
+			break;
+		}
+	}
+	if (format == -1)
+		format = SF_FORMAT_WAV;
+	if (format == SF_FORMAT_WAV && info->channels > 2)
+		format = SF_FORMAT_WAVEX;
+
+	info->format |= format;
+
+	if (format == SF_FORMAT_OGG || format == SF_FORMAT_FLAC)
+		info->format = (info->format & ~SF_FORMAT_ENDMASK) | SF_ENDIAN_FILE;
+	if (format == SF_FORMAT_OGG)
+		info->format = (info->format & ~SF_FORMAT_SUBMASK) | SF_FORMAT_VORBIS;
+}
 
 static int setup_sndfile(struct data *data)
 {
@@ -1354,28 +1245,23 @@ static int setup_sndfile(struct data *data)
 			fprintf(stderr, "error: unknown format \"%s\"\n", data->format);
 			return -EINVAL;
 		}
-		info.format |= SF_FORMAT_WAV;
-#if __BYTE_ORDER == __BIG_ENDIAN
-		info.format |= SF_ENDIAN_BIG;
-#else
-		info.format |= SF_ENDIAN_LITTLE;
-#endif
+		format_from_filename(&info, data->filename);
 	}
 
 	data->file = sf_open(data->filename,
 			data->mode == mode_playback ? SFM_READ : SFM_WRITE,
 			&info);
 	if (!data->file) {
-		fprintf(stderr, "error: failed to open audio file \"%s\": %s\n",
+		fprintf(stderr, "sndfile: failed to open audio file \"%s\": %s\n",
 				data->filename, sf_strerror(NULL));
 		return -EIO;
 	}
 
 	if (data->verbose)
-		printf("opened file \"%s\" format %08x channels:%d rate:%d\n",
+		printf("sndfile: opened file \"%s\" format %08x channels:%d rate:%d\n",
 				data->filename, info.format, info.channels, info.samplerate);
 	if (data->channels > 0 && info.channels != data->channels) {
-		printf("given channels (%u) don't match file channels (%d)\n",
+		fprintf(stderr, "sndfile: given channels (%u) don't match file channels (%d)\n",
 				data->channels, info.channels);
 		return -EINVAL;
 	}
@@ -1399,7 +1285,7 @@ static int setup_sndfile(struct data *data)
 				def = true;
 			}
 			if (data->verbose) {
-				printf("using %s channel map: ", def ? "default" : "file");
+				printf("sndfile: using %s channel map: ", def ? "default" : "file");
 				channelmap_print(&data->channelmap);
 				printf("\n");
 			}
@@ -1462,7 +1348,7 @@ static int setup_sndfile(struct data *data)
 	}
 
 	if (data->verbose)
-		printf("rate=%u channels=%u fmt=%s samplesize=%u stride=%u latency=%u (%.3fs)\n",
+		printf("PCM: rate=%u channels=%u fmt=%s samplesize=%u stride=%u latency=%u (%.3fs)\n",
 				data->rate, data->channels,
 				sf_fmt_to_str(info.format),
 				data->samplesize,
@@ -1523,11 +1409,17 @@ int main(int argc, char *argv[])
 	/* negative means no volume adjustment */
 	data.volume = -1.0;
 	data.quality = -1;
+	data.props = pw_properties_new(
+			PW_KEY_APP_NAME, prog,
+			PW_KEY_NODE_NAME, prog,
+			NULL);
 
-	/* initialize list every time */
-	spa_list_init(&data.targets);
+	if (data.props == NULL) {
+		fprintf(stderr, "error: pw_properties_new() failed: %m\n");
+		goto error_no_props;
+	}
 
-	while ((c = getopt_long(argc, argv, "hvprmdR:q:", long_options, NULL)) != -1) {
+	while ((c = getopt_long(argc, argv, "hvprmdR:q:P:", long_options, NULL)) != -1) {
 
 		switch (c) {
 
@@ -1584,19 +1476,14 @@ int main(int argc, char *argv[])
 			data.media_role = optarg;
 			break;
 
+		case 'P':
+			pw_properties_update_string(data.props, optarg, strlen(optarg));
+			break;
+
 		case OPT_TARGET:
 			data.target = optarg;
-			if (spa_streq(optarg, "auto")) {
-				data.target_id = PW_ID_ANY;
-				break;
-			}
-			if (!isdigit(optarg[0])) {
-				fprintf(stderr, "error: bad target option \"%s\"\n", optarg);
-				goto error_usage;
-			}
-			data.target_id = atoi(optarg);
-			if (data.target_id == 0) {
-				data.target_id = PW_ID_ANY;
+			if (spa_streq(data.target, "0")) {
+				data.target = NULL;
 				flags &= ~PW_STREAM_FLAG_AUTOCONNECT;
 			}
 			break;
@@ -1635,10 +1522,6 @@ int main(int argc, char *argv[])
 			data.volume = atof(optarg);
 			break;
 
-		case OPT_LIST_TARGETS:
-			data.list_targets = true;
-			break;
-
 		default:
 			fprintf(stderr, "error: unknown option '%c'\n", c);
 			goto error_usage;
@@ -1666,10 +1549,7 @@ int main(int argc, char *argv[])
 					DEFAULT_MEDIA_CATEGORY_RECORD;
 	if (!data.media_role)
 		data.media_role = DEFAULT_MEDIA_ROLE;
-	if (!data.target) {
-		data.target = DEFAULT_TARGET;
-		data.target_id = PW_ID_ANY;
-	}
+
 	if (!data.latency)
 		data.latency = data.mode == mode_playback ?
 			DEFAULT_LATENCY_PLAY :
@@ -1690,26 +1570,18 @@ int main(int argc, char *argv[])
 	if (data.volume < 0)
 		data.volume = DEFAULT_VOLUME;
 
-	if (!data.list_targets && optind >= argc) {
-		fprintf(stderr, "error: filename argument missing\n");
+	if (optind >= argc) {
+		fprintf(stderr, "error: filename or - argument missing\n");
 		goto error_usage;
 	}
 	data.filename = argv[optind++];
 
-	data.props = pw_properties_new(
-			PW_KEY_MEDIA_TYPE, data.media_type,
-			PW_KEY_MEDIA_CATEGORY, data.media_category,
-			PW_KEY_MEDIA_ROLE, data.media_role,
-			PW_KEY_APP_NAME, prog,
-			PW_KEY_MEDIA_FILENAME, data.filename,
-			PW_KEY_MEDIA_NAME, data.filename,
-			PW_KEY_NODE_NAME, prog,
-			NULL);
-
-	if (data.props == NULL) {
-		fprintf(stderr, "error: pw_properties_new() failed: %m\n");
-		goto error_no_props;
-	}
+	pw_properties_set(data.props, PW_KEY_MEDIA_TYPE, data.media_type);
+	pw_properties_set(data.props, PW_KEY_MEDIA_CATEGORY, data.media_category);
+	pw_properties_set(data.props, PW_KEY_MEDIA_ROLE, data.media_role);
+	pw_properties_set(data.props, PW_KEY_MEDIA_FILENAME, data.filename);
+	pw_properties_set(data.props, PW_KEY_MEDIA_NAME, data.filename);
+	pw_properties_set(data.props, PW_KEY_TARGET_OBJECT, data.target);
 
 	/* make a main loop. If you already have another main loop, you can add
 	 * the fd of this pipewire mainloop to it. */
@@ -1744,17 +1616,9 @@ int main(int argc, char *argv[])
 	}
 	pw_core_add_listener(data.core, &data.core_listener, &core_events, &data);
 
-	data.registry = pw_core_get_registry(data.core, PW_VERSION_REGISTRY, 0);
-	if (!data.registry) {
-		fprintf(stderr, "error: pw_core_get_registry() failed: %m\n");
-		goto error_no_registry;
-	}
-	pw_registry_add_listener(data.registry, &data.registry_listener, &registry_events, &data);
-
-	data.sync = pw_core_sync(data.core, 0, data.sync);
-
-	if (!data.list_targets) {
-
+	if (spa_streq(data.filename, "-")) {
+		ret = setup_pipe(&data);
+	} else {
 		switch (data.data_type) {
 		case TYPE_PCM:
 			ret = setup_sndfile(&data);
@@ -1769,105 +1633,102 @@ int main(int argc, char *argv[])
 			ret = -ENOTSUP;
 			break;
 		}
+	}
 
-		if (ret < 0) {
-			fprintf(stderr, "error: open failed: %s\n", spa_strerror(ret));
-			switch (ret) {
-			case -EIO:
-				goto error_bad_file;
-			case -EINVAL:
-			default:
-				goto error_usage;
-			}
+	if (ret < 0) {
+		fprintf(stderr, "error: open failed: %s\n", spa_strerror(ret));
+		switch (ret) {
+		case -EIO:
+			goto error_bad_file;
+		case -EINVAL:
+		default:
+			goto error_usage;
 		}
-		switch (data.data_type) {
-		case TYPE_PCM:
-		{
-			struct spa_audio_info_raw info;
-			info = SPA_AUDIO_INFO_RAW_INIT(
-				.flags = data.channelmap.n_channels ? 0 : SPA_AUDIO_FLAG_UNPOSITIONED,
-				.format = data.spa_format,
-				.rate = data.rate,
-				.channels = data.channels);
+	}
+	switch (data.data_type) {
+	case TYPE_PCM:
+	{
+		struct spa_audio_info_raw info;
+		info = SPA_AUDIO_INFO_RAW_INIT(
+			.flags = data.channelmap.n_channels ? 0 : SPA_AUDIO_FLAG_UNPOSITIONED,
+			.format = data.spa_format,
+			.rate = data.rate,
+			.channels = data.channels);
 
-			if (data.channelmap.n_channels)
-				memcpy(info.position, data.channelmap.channels, data.channels * sizeof(int));
+		if (data.channelmap.n_channels)
+			memcpy(info.position, data.channelmap.channels, data.channels * sizeof(int));
 
-			params[0] = spa_format_audio_raw_build(&b, SPA_PARAM_EnumFormat, &info);
-			break;
+		params[0] = spa_format_audio_raw_build(&b, SPA_PARAM_EnumFormat, &info);
+		break;
+	}
+	case TYPE_MIDI:
+		params[0] = spa_pod_builder_add_object(&b,
+				SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat,
+				SPA_FORMAT_mediaType,		SPA_POD_Id(SPA_MEDIA_TYPE_application),
+				SPA_FORMAT_mediaSubtype,	SPA_POD_Id(SPA_MEDIA_SUBTYPE_control));
+
+		pw_properties_set(data.props, PW_KEY_FORMAT_DSP, "8 bit raw midi");
+		break;
+	case TYPE_DSD:
+	{
+		struct spa_audio_info_dsd info;
+		size_t i;
+
+		spa_zero(info);
+		info.channels = data.dsf.info.channels;
+		info.rate = data.dsf.info.rate / 8;
+
+		for (i = 0; i < SPA_N_ELEMENTS(dsd_layouts); i++) {
+			if (dsd_layouts[i].type != data.dsf.info.channel_type)
+				continue;
+			info.channels = dsd_layouts[i].info.n_channels;
+			memcpy(info.position, dsd_layouts[i].info.position,
+					info.channels * sizeof(uint32_t));
 		}
-		case TYPE_MIDI:
-			params[0] = spa_pod_builder_add_object(&b,
-					SPA_TYPE_OBJECT_Format, SPA_PARAM_EnumFormat,
-					SPA_FORMAT_mediaType,		SPA_POD_Id(SPA_MEDIA_TYPE_application),
-					SPA_FORMAT_mediaSubtype,	SPA_POD_Id(SPA_MEDIA_SUBTYPE_control));
+		params[0] = spa_format_audio_dsd_build(&b, SPA_PARAM_EnumFormat, &info);
+		break;
+	}
+	}
 
-			pw_properties_set(data.props, PW_KEY_FORMAT_DSP, "8 bit raw midi");
-			break;
-		case TYPE_DSD:
-		{
-			struct spa_audio_info_dsd info;
-			size_t i;
+	data.stream = pw_stream_new(data.core, prog, data.props);
+	data.props = NULL;
 
-			spa_zero(info);
-			info.channels = data.dsf.info.channels;
-			info.rate = data.dsf.info.rate / 8;
+	if (data.stream == NULL) {
+		fprintf(stderr, "error: failed to create stream: %m\n");
+		goto error_no_stream;
+	}
+	pw_stream_add_listener(data.stream, &data.stream_listener, &stream_events, &data);
 
-			for (i = 0; i < SPA_N_ELEMENTS(dsd_layouts); i++) {
-				if (dsd_layouts[i].type != data.dsf.info.channel_type)
-					continue;
-				info.channels = dsd_layouts[i].info.n_channels;
-				memcpy(info.position, dsd_layouts[i].info.position,
-						info.channels * sizeof(uint32_t));
-			}
-			params[0] = spa_format_audio_dsd_build(&b, SPA_PARAM_EnumFormat, &info);
-			break;
-		}
-		}
+	if (data.verbose)
+		printf("connecting %s stream; target=%s\n",
+				data.mode == mode_playback ? "playback" : "record",
+				data.target);
 
-		data.stream = pw_stream_new(data.core, prog, data.props);
-		data.props = NULL;
+	if (data.verbose)
+		data.timer = pw_loop_add_timer(l, do_print_delay, &data);
 
-		if (data.stream == NULL) {
-			fprintf(stderr, "error: failed to create stream: %m\n");
-			goto error_no_stream;
-		}
-		pw_stream_add_listener(data.stream, &data.stream_listener, &stream_events, &data);
+	ret = pw_stream_connect(data.stream,
+			  data.mode == mode_playback ? PW_DIRECTION_OUTPUT : PW_DIRECTION_INPUT,
+			  PW_ID_ANY,
+			  flags |
+			  PW_STREAM_FLAG_MAP_BUFFERS,
+			  params, 1);
+	if (ret < 0) {
+		fprintf(stderr, "error: failed connect: %s\n", spa_strerror(ret));
+		goto error_connect_fail;
+	}
 
-		if (data.verbose)
-			printf("connecting %s stream; target_id=%"PRIu32"\n",
-					data.mode == mode_playback ? "playback" : "record",
-					data.target_id);
+	if (data.verbose) {
+		const struct pw_properties *props;
+		void *pstate;
+		const char *key, *val;
 
-		if (data.verbose) {
-			struct timespec timeout = {0, 1}, interval = {1, 0};
-			struct spa_source *timer = pw_loop_add_timer(l, do_print_delay, &data);
-			pw_loop_update_timer(l, timer, &timeout, &interval, false);
-		}
-
-		ret = pw_stream_connect(data.stream,
-				  data.mode == mode_playback ? PW_DIRECTION_OUTPUT : PW_DIRECTION_INPUT,
-				  data.target_id,
-				  flags |
-				  PW_STREAM_FLAG_MAP_BUFFERS,
-				  params, 1);
-		if (ret < 0) {
-			fprintf(stderr, "error: failed connect: %s\n", spa_strerror(ret));
-			goto error_connect_fail;
-		}
-
-		if (data.verbose) {
-			const struct pw_properties *props;
-			void *pstate;
-			const char *key, *val;
-
-			if ((props = pw_stream_get_properties(data.stream)) != NULL) {
-				printf("stream properties:\n");
-				pstate = NULL;
-				while ((key = pw_properties_iterate(props, &pstate)) != NULL &&
-					(val = pw_properties_get(props, key)) != NULL) {
-					printf("\t%s = \"%s\"\n", key, val);
-				}
+		if ((props = pw_stream_get_properties(data.stream)) != NULL) {
+			printf("stream properties:\n");
+			pstate = NULL;
+			while ((key = pw_properties_iterate(props, &pstate)) != NULL &&
+				(val = pw_properties_get(props, key)) != NULL) {
+				printf("\t%s = \"%s\"\n", key, val);
 			}
 		}
 	}
@@ -1876,61 +1737,8 @@ int main(int argc, char *argv[])
 	pw_main_loop_run(data.loop);
 
 	/* we're returning OK only if got to the point to drain */
-	if (!data.list_targets) {
-		if (data.drained)
-			exit_code = EXIT_SUCCESS;
-	} else {
-		if (data.targets_listed) {
-			struct target *target, *target_default;
-			char *default_name;
-
-			default_name = (data.mode == mode_record) ?
-				data.default_source : data.default_sink;
-
-			exit_code = EXIT_SUCCESS;
-
-			/* first find the highest priority */
-			target_default = NULL;
-			spa_list_for_each(target, &data.targets, link) {
-				if (target_default == NULL ||
-				    spa_streq(default_name, target->name) ||
-				    (default_name[0] == '\0' &&
-				     target->prio > target_default->prio))
-					target_default = target;
-			}
-			printf("Available targets (\"*\" denotes default): %s\n", default_name);
-			spa_list_for_each(target, &data.targets, link) {
-				if (target->type != TARGET_TYPE_SOURCE)
-					continue;
-				printf("%s\t%"PRIu32": source description=\"%s\" prio=%d\n",
-				       target == target_default ? "*" : "",
-				       target->id, target->desc, target->prio);
-			}
-			spa_list_for_each(target, &data.targets, link) {
-				if (target->type != TARGET_TYPE_SINK)
-					continue;
-				printf("%s\t%"PRIu32": %s description=\"%s\" prio=%d\n",
-					target == target_default ? "*" : "",
-					target->id, data.mode == mode_record ? "monitor" : "sink",
-					target->desc, target->prio);
-			}
-			spa_list_for_each(target, &data.targets, link) {
-				if (target->type != TARGET_TYPE_STREAM)
-					continue;
-				printf("%s\t%"PRIu32": stream description=\"%s\" prio=%d\n",
-				       target == target_default ? "*" : "",
-				       target->id, target->desc, target->prio);
-			}
-		}
-	}
-
-	/* destroy targets */
-	while (!spa_list_is_empty(&data.targets)) {
-		struct target *target;
-		target = spa_list_last(&data.targets, struct target, link);
-		spa_list_remove(&target->link);
-		target_destroy(target);
-	}
+	if (data.drained)
+		exit_code = EXIT_SUCCESS;
 
 error_connect_fail:
 	if (data.stream) {
@@ -1938,15 +1746,6 @@ error_connect_fail:
 		pw_stream_destroy(data.stream);
 	}
 error_no_stream:
-	if (data.metadata) {
-		spa_hook_remove(&data.metadata_listener);
-		pw_proxy_destroy((struct pw_proxy*)data.metadata);
-	}
-	if (data.registry) {
-		spa_hook_remove(&data.registry_listener);
-		pw_proxy_destroy((struct pw_proxy*)data.registry);
-	}
-error_no_registry:
 	spa_hook_remove(&data.core_listener);
 	pw_core_disconnect(data.core);
 error_ctx_connect_failed:
